#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"../port/error.h"
#include "io.h"

/*
 * mk48txx real time clock and non-volatile ram
 */

enum {
	Epoch=		1970,		/* system epoch */
	EpochDay=	4,			/* jan1 = thurs */
	Year0=		1968,		/* rtc epoch */

	/* mk48t59 - other models will be different */
	Nvsize=		8192,
	ClkOff=		Nvsize-16,

	/* Csr bits */
	Read=		1<<6,
	Write=		1<<7,
};

typedef struct Mk48txx Mk48txx;
struct Mk48txx {
	char		nvram[ClkOff];
	char		pad0[8];
	char		csr;

	/* these are in BCD */
	char		sec;
	char		min;
	char		hour;
	char		wday;
	char		mday;
	char		mon;
	char		year;
};

typedef struct Ctlr Ctlr;
struct Ctlr {
	Mk48txx	*dev;
	Lock;
};

typedef struct Rtc	Rtc;
struct Rtc
{
	int	sec;
	int	min;
	int	hour;
	int	wday;
	int	mday;
	int	mon;
	int	year;
};

enum {
	Qdir = 0,
	Qrtc,
	Qnvram,
};

static Dirtab rtcdir[]={
	".",		{Qdir, 0, QTDIR},	0,	0555,
	"nvram",	{Qnvram, 0},	Nvsize,	0664,
	"rtc",		{Qrtc, 0},	0,	0664,
};

static Ctlr ctlr0;

static ulong rtc2sec(Rtc*);
static void sec2rtc(ulong, Rtc*);

static int
debcd(uchar c)
{
	return (c&0xf) + 10*(c>>4);
}

static uchar
enbcd(int n)
{
	return (n%10) | (((n/10) % 10)<<4);
}

/* connect the rtc thats on the sbus */
static void
rtcsbus(void)
{
	KMap *k;

	k = kmapphys(EEPROM, 1, PTEIO|PTEUNCACHED|PTELOCK);
	ctlr0.dev = (void*)VA(k);
}

static Chan*
rtcattach(char* spec)
{
	return devattach('r', spec);
}

static Walkqid*	 
rtcwalk(Chan* c, Chan *nc, char** name, int nname)
{
	return devwalk(c, nc, name, nname, rtcdir, nelem(rtcdir), devgen);
}

static int	 
rtcstat(Chan* c, uchar* dp, int n)
{
	return devstat(c, dp, n, rtcdir, nelem(rtcdir), devgen);
}

static Chan*
rtcopen(Chan* c, int omode)
{
	omode = openmode(omode);
	switch((ulong)c->qid.path){
	case Qrtc:
		if(strcmp(up->user, eve)!=0 && omode!=OREAD)
			error(Eperm);
		break;
	case Qnvram:
		if(strcmp(up->user, eve)!=0)
			error(Eperm);
	}
	return devopen(c, omode, rtcdir, nelem(rtcdir), devgen);
}

static void	 
rtcclose(Chan*)
{
}


/* called with lock held */
static long	 
_rtctime(Ctlr *ctlr)
{
	Rtc rtc;
	Mk48txx *dev;

	ilock(ctlr);
	dev = ctlr->dev;
	dev->csr |= Read;
	rtc.sec = debcd(dev->sec);
	rtc.min = debcd(dev->min);
	rtc.hour = debcd(dev->hour);
	rtc.wday = dev->wday;
	rtc.mday = debcd(dev->mday);
	rtc.mon = debcd(dev->mon);
	rtc.year = debcd(dev->year);
	dev->csr &= ~Read;
	iunlock(ctlr);
	rtc.year += Year0;
	if(rtc.year < Year0)
		rtc.year += 100;
	return rtc2sec(&rtc);
}

static void
rtcset(Ctlr *ctlr, ulong secs)
{
	Rtc rtc;
	Mk48txx *dev;

	sec2rtc(secs, &rtc);
	rtc.year -= Year0;
	if(rtc.year < 0)
		rtc.year += 100;

	ilock(ctlr);
	dev = ctlr->dev;
	dev->csr |= Write;
	dev->sec = enbcd(rtc.sec);
	dev->min = enbcd(rtc.min);
	dev->hour = enbcd(rtc.hour);
	dev->wday = rtc.wday + 1;
	dev->mday = enbcd(rtc.mday);
	dev->mon = enbcd(rtc.mon);
	dev->year = enbcd(rtc.year);
	dev->csr &= ~Write;
	iunlock(ctlr);
}

static long	 
rtcread(Chan* c, void* buf, long n, vlong off)
{
	ulong t;
	ulong offset = off;

	if(c->qid.type & QTDIR)
		return devdirread(c, buf, n, rtcdir, nelem(rtcdir), devgen);

	switch((ulong)c->qid.path){
	case Qrtc:
		t = _rtctime(&ctlr0);
		n = readnum(offset, buf, n, t, 12);
		return n;
	case Qnvram:
		/* currently unsupported */
		break;
	}
	error(Ebadarg);
	return 0;
}


static long	 
rtcwrite(Chan* c, void* buf, long n, vlong off)
{
	ulong secs;
	char *cp, *ep;
	ulong offset = off;

	if(offset!=0)
		error(Ebadarg);


	switch((ulong)c->qid.path){
	case Qrtc:
		/*
		 *  read the time
		 */
		cp = ep = buf;
		ep += n;
		while(cp < ep){
			if(*cp>='0' && *cp<='9')
				break;
			cp++;
		}
		secs = strtoul(cp, 0, 0);
		rtcset(&ctlr0, secs);
		return n;
	case Qnvram:
		/* currently unsupported */
		break;
	}
	error(Ebadarg);
	return 0;
}

Dev rtcdevtab = {
	'r',
	"rtc",

	devreset,
	rtcsbus,
	devshutdown,
	rtcattach,
	rtcwalk,
	rtcstat,
	rtcopen,
	devcreate,
	rtcclose,
	rtcread,
	devbread,
	rtcwrite,
	devbwrite,
	devremove,
	devwstat,
};

#define SEC2MIN 60L
#define SEC2HOUR (60L*SEC2MIN)
#define SEC2DAY (24L*SEC2HOUR)

/*
 *  days per month plus days/year
 */
static	int	dmsize[] =
{
	365, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};
static	int	ldmsize[] =
{
	366, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

/*
 *  return the days/month for the given year
 */
static int*
yrsize(int y)
{
	if((y%4) == 0 && ((y%100) != 0 || (y%400) == 0))
		return ldmsize;
	else
		return dmsize;
}

/*
 *  compute seconds since Epoch
 *  wday ignored.
 */
static ulong
rtc2sec(Rtc *rtc)
{
	ulong secs;
	int i;
	int *d2m;

	secs = 0;

	/*
	 *  seconds per year
	 */
	for(i = Epoch; i < rtc->year; i++){
		d2m = yrsize(i);
		secs += d2m[0] * SEC2DAY;
	}

	/*
	 *  seconds per month
	 */
	d2m = yrsize(rtc->year);
	for(i = 1; i < rtc->mon; i++)
		secs += d2m[i] * SEC2DAY;

	secs += (rtc->mday-1) * SEC2DAY;
	secs += rtc->hour * SEC2HOUR;
	secs += rtc->min * SEC2MIN;
	secs += rtc->sec;

	return secs;
}

/*
 *  compute rtc from seconds since Epoch
 */
static void
sec2rtc(ulong secs, Rtc *rtc)
{
	int d;
	long hms, day;
	int *d2m;

	/*
	 * break initial number into days
	 */
	hms = secs % SEC2DAY;
	day = secs / SEC2DAY;
	if(hms < 0) {
		hms += SEC2DAY;
		day -= 1;
	}

	/*
	 * generate hours:minutes:seconds
	 */
	rtc->sec = hms % 60;
	d = hms / 60;
	rtc->min = d % 60;
	d /= 60;
	rtc->hour = d;

	rtc->wday = (day + EpochDay) % 7;

	/*
	 * year number
	 */
	if(day >= 0)
		for(d = Epoch; day >= *yrsize(d); d++)
			day -= *yrsize(d);
	else
		for (d = Epoch; day < 0; d--)
			day += *yrsize(d-1);
	rtc->year = d;

	/*
	 * generate month
	 */
	d2m = yrsize(rtc->year);
	for(d = 1; day >= d2m[d]; d++)
		day -= d2m[d];
	rtc->mday = day + 1;
	rtc->mon = d;

	return;
}

