/*
 * ethernet driver for happy meal ethernet on sbus.
 * Uses the ethermii to control PHY.
 * built on the generic devether framework.
 *
 * XXX Does not have a mechanism for selecting external AUI yet.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/netif.h"

#include "etherif.h"
#include "ethermii.h"

enum {
	KB = 1024,
	NDesc = 32,				/* 32,64,128,256 */
	/* XXX with this setting I get PackOv's for packets as small as 1450, why? */
	//EthBufSz = ROUNDUP(ETHERMAXTU + 4, 64),
	EthBufSz = 2048,

	/* descriptor flags */
	PackEnd = 1<<29,
	PackStart = 1<<30,
	PackOv = 1<<30,
	OwnHw = 1<<31,			/* descriptor owned by hardware */

	/* registers */
	GlobalReset = 0x0,
	GlobalConfig = 0x4,
	GlobalStatus = 0x100,
	GlobalIntMask = 0x104,
	TxPending = 0x2000,
	TxConfig = 0x2004,
	TxDescBase = 0x2008,
	TxDescSize = 0x202c,
	RxConfig = 0x4000,
	RxDescBase = 0x4004,
	RxFifoCount = 0x4018,
	XifConfig = 0x6000,
	TxMacConfig = 0x620c,
	TxMaxFrame = 0x6230,
	TxMinFrame = 0x6234,
	CollideCount = 0x6240,		/* 4 counters */
	RandSeed = 0x6250,
	RxMacConfig = 0x630c,
	RxMaxFrame = 0x6310,
	RxMinFrame = 0x6314,
	MacAddr = 0x6320,
	RxCount = 0x6324,
	BadLenCount = 0x6328,
	BadAlignCount = 0x632c,
	BadFcsCount = 0x6330,
	BadCodeCount = 0x6338,
	MifFrame = 0x700c,
	MifConfig = 0x7010,

	/* GlobalIntMask bits */
	IntRxDone = 1<<16,
	IntRxNoBuf = 1<<17,
	IntMIF = 1<<23,
	IntTxDone = 1<<24,
	IntTxAll = 1<<25,
	IntError = 0xfc7c0620,	/* bits 5,9,10,18..22,26..31 */

	/* TxConfig bits */
	TxDmaEnable = 1<<0,

	/* RxConfig bits */
	RxDmaEnable = 1<<0,

	/* XifConfig bits */
	TxEnable = 1<<0,
	MiiEnable = 1<<3,

	/* MifConfig bits */
	AltPhy = 1<<0,
	BitBangMode = 1<<2,

	/* TxMacConfig bits */
	TxMacEnable = 1<<0,
	TxFullDuplex = 1<<9,
	TxNeverGiveUp = 1<<10,

	/* RxMacConfig bits */
	RxMacEnable = 1<<0,
	RxStripPad = 1<<5,
	RxPromisc = 1<<6,
	RxRejectMine = 1<<9,
	RxGroupPromisc = 1<<10,

	/* MifFrame bits */
	FrameWrite = (1<<30)|(1<<28)|(1<<17),
	FrameRead = (1<<30)|(2<<28)|(1<<17),
	FrameDone = 1<<16,
};

typedef struct Desc Desc;
struct Desc {
	ulong	flags;
	ulong	data;
};

typedef struct Ctlr Ctlr;
struct Ctlr {
	Lock;
	Ether*	ether;
	ulong	base;			/* mapped registers */

	Mii		mii;
	Desc*	rd;				/* descriptor rings for read and write */
	Block*	rb[NDesc];
	int		rdh;
	Desc*	td;
	Block*	tb[NDesc];
	int		tdh;

	ulong	ncollisions;		/* stats */
	ulong	nrxframes;
	ulong	ntxframes;
	ulong	nbadframes;

	uchar	ea[Eaddrlen];
};


/* just one */
static Ctlr *ctlr0 = 0;

static void
reg32w(Ctlr *c, ulong addr, uint val)
{
	*(uint*)(c->base + addr) = val;
}

static uint
reg32r(Ctlr *c, ulong addr)
{
	return *(uint*)(c->base + addr);
}

static int
miiread(Mii *mii, int phy, int reg)
{
	Ctlr *ctlr;
	int v;

	ctlr = mii->ctlr;
	ilock(ctlr);
	reg32w(ctlr, MifFrame, FrameRead | (phy<<23) | (reg<<18));
	while(((v = reg32r(ctlr, MifFrame)) & FrameDone) == 0)
		continue;
	iunlock(ctlr);
	return v & 0xffff;
}

static int
miiwrite(Mii *mii, int phy, int reg, int data)
{
	Ctlr *ctlr;

	ctlr = mii->ctlr;
	ilock(ctlr);
	reg32w(ctlr, MifFrame, FrameWrite | (phy<<23) | (reg<<18) | data);
	while((reg32r(ctlr, MifFrame) & FrameDone) == 0)
		continue;
	iunlock(ctlr);
	return 0;
}

/* allocate a block for receiving, ensuring 64-byte alignment */
static Block *
iallocbRCV(void)
{
	Block *bp;

	bp = iallocb(EthBufSz + 64);
	if(bp)
		bp->rp = (uchar*)ROUNDUP((ulong)bp->rp, 64);
	return bp;
}

static void
hmepromiscuous(void* arg, int on)
{
	Ether* ether;
	Ctlr* ctlr;
	ulong val;

	/* XXX is this a bad way to poll? 
	 * if we yield here, we may have the mac disabled for a bit of time,
	 * however, if we spin here, we're needlessly using cpu that could be
	 * better used by others.
	 */
	ether = arg;
	ctlr = ether->ctlr;
	ilock(ctlr);
	val = reg32r(ctlr, RxMacConfig) & ~RxPromisc;
	reg32w(ctlr, RxMacConfig, 0);
	while(reg32r(ctlr, RxMacConfig) & 1)
		continue;

	if(on)
		val |= RxPromisc;
	reg32w(ctlr, RxMacConfig, val);
	iunlock(ctlr);
}

static long
hmeifstat(Ether* ether, void* a, long n, ulong offset)
{
	Ctlr* ctlr;
	char* p;
	int len;

	ctlr = ether->ctlr;
	if(n == 0)
		return 0;

	p = malloc(READSTR);
	len = snprint(p, READSTR, "RxFrames: %lud\n", ctlr->nrxframes);
	len += snprint(p+len, READSTR-len, "TxFrames: %lud\n", ctlr->ntxframes);
	len += snprint(p+len, READSTR-len, "BadFrames: %lud\n", ctlr->nbadframes);
	len += snprint(p+len, READSTR-len, "Collisions: %lud\n", ctlr->ncollisions);
	USED(len);

	n = readstr(offset, a, n, p);
	free(p);

	return n;	
}

static void
hmeattach(Ether *ether)
{
	/* we've already done all the work */
	USED(ether);
}

/* 
 * Queue packets into the transmit ring.
 * Called with controller locked.
 */
static void
hmetxstart(Ether* ether)
{
	Ctlr* ctlr;
	Desc* td;
	Block *bp, **tb;
	int h, len;

	ctlr = ether->ctlr;
	td = ctlr->td;
	tb = ctlr->tb;
	for(h = ctlr->tdh; !(td[h].flags & OwnHw); h = NEXT(h, NDesc)) {
		bp = qget(ether->oq);
		if(!bp)
			break;
		len = BLEN(bp);
		if(len > EthBufSz)
			panic("buffer too large for device");

		if(tb[h])
			freeb(tb[h]);
		tb[h] = bp;
		td[h].data = PADDR(bp->rp);
		td[h].flags = len | PackStart | PackEnd | OwnHw;
		coherence();
		reg32w(ctlr, TxPending, 1);
		ctlr->ntxframes ++;
	}
	ctlr->tdh = h;
}

static void
hmetransmit(Ether* ether)
{
	Ctlr* ctlr;

	ctlr = ether->ctlr;
	ilock(ctlr);
	hmetxstart(ether);
	iunlock(ctlr);
}

/*
 * grab items from the receive ring and pass them up.
 * called from interrupt.
 */
static void
hmereceive(Ether* ether)
{
	Ctlr* ctlr;
	Desc* rd;
	Block *bp, **rb;
	int h, len;

	ctlr = ether->ctlr;
	rd = ctlr->rd;
	rb = ctlr->rb;
	for(h = ctlr->rdh; !(rd[h].flags & OwnHw); h = NEXT(h, NDesc)) {
		len = (rd[h].flags >> 16) & 0x1fff;
		if(rd[h].flags & PackOv) {
			print("#l%d: overflow %d\n", ether->ctlrno, len);
		} else {
			bp = iallocbRCV();
			if(bp) {
				rb[h]->wp = rb[h]->rp + len;
				coherence();
				etheriq(ether, rb[h], 1);
				rb[h] = bp;
				rd[h].data = PADDR(bp->rp);
			} /* otherwise ignore data and reuse buffer */
		}
		rd[h].flags = OwnHw | (EthBufSz<<16);
		coherence();
	}
	ctlr->rdh = h;
}

static void
hmestats(Ctlr* ctlr)
{
	ctlr->ncollisions += reg32r(ctlr, CollideCount);
	reg32w(ctlr, CollideCount, 0);
	ctlr->nrxframes += reg32r(ctlr, RxCount);
	reg32w(ctlr, RxCount, 0);
	ctlr->nbadframes += reg32r(ctlr, BadLenCount) + reg32r(ctlr, BadAlignCount) +
					  reg32r(ctlr, BadFcsCount) + reg32r(ctlr, BadCodeCount);
	reg32w(ctlr, BadLenCount, 0);
	reg32w(ctlr, BadFcsCount, 0);
	reg32w(ctlr, BadAlignCount, 0);
	reg32w(ctlr, BadCodeCount, 0);
}

void
hmeintr(void)
{
	Ctlr* ctlr;
	Ether* ether;
	ulong status;

	ctlr = ctlr0;
	ether = ctlr->ether;
	status = reg32r(ctlr, GlobalStatus);
	hmestats(ctlr);
	if(status & (IntRxDone | IntRxNoBuf)) {
		ilock(ctlr);
		hmereceive(ether);
		iunlock(ctlr);
	}

	if(status & (IntTxDone | IntTxAll)) {
		ilock(ctlr);
		hmetxstart(ether);
		iunlock(ctlr);
	}

	if(status & IntError) {
		print("#l%d: error status %lux\n", ether->ctlrno, status);
	}
}

/* attach hme from the sbus/feps */
static void
hmesbus(void)
{
	KMap *k;
	Ctlr *ctlr;
	Block *bp;
	int i;

	k = kmapphys(HME, 4, PTEIO|PTEUNCACHED|PTELOCK);
	ctlr = malloc(sizeof *ctlr);
	memset(ctlr, 0, sizeof *ctlr);
	ctlr->base = VA(k);
	ctlr->mii.mir = miiread;
	ctlr->mii.miw = miiwrite;
	ctlr->mii.ctlr = ctlr;
	reg32w(ctlr, GlobalReset, 3);
	while(reg32r(ctlr, GlobalReset)  != 0)
		continue;

	ctlr->rd = mallocalign(NDesc*sizeof *ctlr->rd, 2*KB, 0, 0);
	ctlr->td = mallocalign(NDesc*sizeof *ctlr->td, 2*KB, 0, 0);
	memmove(ctlr->ea, conf.idprom + 2, Eaddrlen);

	for(i = 0; i < NDesc; i++) {
		bp = iallocbRCV();
		if(!bp)
			panic("can't allocate ethernet receive ring\n");
		ctlr->rb[i] = bp;
		ctlr->rd[i].data = PADDR(bp->rp);
		ctlr->rd[i].flags = OwnHw | (EthBufSz<<16);
		ctlr->tb[i] = 0;
		ctlr->td[i].data = 0;
		ctlr->td[i].flags = 0;
	}

	sbusenabledma(SbusFeps);
	sbusenableintr(SbusIntrMapHme);

	ctlr0 = ctlr;
}

static void
hmeinit(Ether *ether)
{
	MiiPhy *phy;
	Ctlr* ctlr;
	int ringsize;

	ctlr = ether->ctlr;
	reg32w(ctlr, GlobalReset, 3);
	while(reg32r(ctlr, GlobalReset)  != 0)
		continue;
	reg32w(ctlr, MifConfig, 0); /* no bitbang, no altphys */

	reg32w(ctlr, RandSeed, ((ether->ea[4]<<8)|ether->ea[5]) & 0x3ff);
	reg32w(ctlr, MacAddr, (ether->ea[0]<<8) | ether->ea[1]);
	reg32w(ctlr, MacAddr-4, (ether->ea[2]<<8) | ether->ea[3]);
	reg32w(ctlr, MacAddr-8, (ether->ea[4]<<8) | ether->ea[5]);

	reg32w(ctlr, TxDescBase, PADDR(ctlr->td));
	reg32w(ctlr, TxDescSize, (NDesc >> 4) - 1);
	reg32w(ctlr, RxDescBase, PADDR(ctlr->rd));

	reg32w(ctlr, GlobalIntMask, IntRxDone | IntRxNoBuf | IntTxDone | IntTxAll | IntError);

	/* note: freebsd src says using 64byte dma on pci can cause hangs */
	reg32w(ctlr, GlobalConfig, 0xe);	/* 64byte dma with parity */
	reg32w(ctlr, TxConfig, reg32r(ctlr, TxConfig) | TxDmaEnable);

	switch(NDesc) {
	case 32:	ringsize = 0; break;
	case 64:	ringsize = 1; break;
	case 128:	ringsize = 2; break;
	case 256:	ringsize = 3; break;
	default: 
		ringsize = 0;
		panic("bad NDesc");
	}
	reg32w(ctlr, RxConfig, RxDmaEnable | (ringsize<<9));
	reg32w(ctlr, XifConfig, TxEnable); /* XXX MiiEnable if external xcvr */

	reg32w(ctlr, TxMacConfig, TxMacEnable | TxFullDuplex | TxNeverGiveUp);
	reg32w(ctlr, TxMaxFrame, ether->maxmtu + 4);
	reg32w(ctlr, TxMinFrame, ether->minmtu);
	reg32w(ctlr, RxMaxFrame, ether->maxmtu + 4);
	reg32w(ctlr, RxMinFrame, ether->minmtu);
	reg32w(ctlr, RxMacConfig, RxMacEnable | RxStripPad);

	/* direct the PHY to autonegotiate */
	if(mii(&ctlr->mii, ~0) == 0 || (phy = ctlr->mii.curphy) == nil) {
		print("hme: init mii failure\n");
		return;
	}
	miistatus(&ctlr->mii);
	ether->mbps = phy->speed;
	print("oui %X phyno %d speed %d %s\n", phy->oui, phy->phyno, 
			phy->speed, phy->fd ? "full duplex": "");
}

static int
hmereset(Ether *ether)
{
	static int scanned = 0;
	Ctlr* ctlr;

	if(scanned)
		return -1;

	hmesbus();
	scanned = 1;

	ctlr = ctlr0;
	ctlr->ether = ether;
	ether->ctlr = ctlr;
	ether->attach = hmeattach;
	ether->transmit = hmetransmit;
	ether->ifstat = hmeifstat;
	ether->promiscuous = hmepromiscuous;
	ether->multicast = 0; /*XXX*/
	ether->shutdown = 0; /*XXX*/
	ether->arg = ether;
	ether->mbps = 100;
	memmove(ether->ea, ctlr->ea, Eaddrlen);
	hmeinit(ether);
	return 0;
}

void
etherhmelink(void)
{
	addethercard("HME", hmereset);
}
