
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "ureg.h"

/* globals */
Mach mach0;
FPsave initfp;
Conf conf;	/* port */



enum {
	/* lsu bits */
	Icache = 1<<0,
	Dcache = 1<<1,
	Immu = 1<<2,
	Dmmu = 1<<3,
};


/* port */
void
reboot(void *entry, void *code, ulong size)
{
	USED(entry, code, size);
	exit(0);
}

static void
parseprops(void)
{
	static char bootarg[256];
	char *p;
	int delim;

	conf.nprops = 0;
	if(ofgetdevprop("/chosen", "bootargs", bootarg, sizeof bootarg) <= 0)
		return;
	bootarg[sizeof bootarg - 1] = '\0';

	/* 
	 * Parse out "word=val" or simply "word" options
	 * from the command line.  Val may be quoted.
	 * note: tabs are not considered delimiters.
	 */
	for(p = bootarg; *p && conf.nprops < NPROPS; ) {
		while(*p == ' ')
			p++;
		if(!*p)
			break;
		conf.propname[conf.nprops] = p;
		while(*p && *p != ' ' && *p != '=')
			*p++;
		if(*p == '=') {
			*p++ = '\0';
			if(*p == '"' || *p == '\'')
				delim = *p++;
			else
				delim = ' ';
			conf.propvalue[conf.nprops] = p;
			while(*p && *p != delim)
				p++;
		} else
			conf.propvalue[conf.nprops] = "";
		if(*p)
			*p++ = '\0';
		conf.nprops++;	
	}
}

/* get a configuration property */
char*
getconf(char* prop)
{
	int i;

	for(i = 0; i < conf.nprops; i++) {
		if(strcmp(conf.propname[i], prop) == 0)
			return conf.propvalue[i];
	}
	return 0;
}

/* declared in port but only used here. */
void
confinit(void)
{
	struct {
		uvlong	pa;
		uvlong	sz;
	} ram[32];
	ulong ktop, nram, i, len;

	/* XXX panic isnt so useful yet */
	if(ofgetdevprop("/SUNW,UltraSPARC", "clock-frequency", &conf.cpufreq, sizeof conf.cpufreq) != sizeof conf.cpufreq)
		panic("cant get cpufreq");
	if(ofgetdevprop("/", "idprom", conf.idprom, sizeof conf.idprom) < sizeof conf.idprom)
		panic("cant get idprom");
	len = ofgetdevprop("/memory", "reg", ram, sizeof ram);
	if(len <= 0 || (len % sizeof ram[0]) != 0)
		panic("cant get memory");
	nram = len / sizeof ram[0];
	parseprops();

	/* 
	 * we can only use two ram regions, and only
	 * if they can map linearly into KZERO.
	 */
	len = 0;
	conf.base1 = 0;
	conf.npage1 = 0;
	for(i = 0; i < nram; i++) {
		if(len >= 2 || ram[i].pa >= RAMTOP) {
			xprint("Skipping all %dKB in bank%d at %x\n", 
				(ulong)ram[i].sz/1024, i, (ulong)ram[i].pa);
			continue;
		}
		if(ram[i].pa + ram[i].sz > RAMTOP) {
			xprint("Skipping %dKB in bank%d at %x\n", 
				((ulong)ram[i].sz - RAMTOP)/1024, i, (ulong)ram[i].pa);
			ram[i].sz = RAMTOP - ram[i].pa;
		}
		if(len == 0) {
			conf.base0 = ram[i].pa;
			conf.npage0 = ram[i].sz / BY2PG;
		} else {
			conf.base1 = ram[i].pa;
			conf.npage1 = ram[i].sz / BY2PG;
		}
		len++;
	}

	ktop = PGROUND((ulong)end);
	ktop = PADDR(ktop);

	/* sanity -- this shouldn't happen */
	if(len == 0 || conf.base0 != 0 || conf.npage0 < ktop/BY2PG) {
		xprint("No suitable RAM found\n");
		panic("No suitable RAM found");
	}

	conf.base0 += ktop;
	conf.npage0 -= ktop/BY2PG;
	conf.npage = conf.npage0 + conf.npage1;

	/* XXX tweak this later */
	conf.nmach = 1;
	conf.monitor = 0;
	conf.nimage = 100;
	conf.nproc = 200;

	/* XXX hmm.. 30% seems excessive for large rams, no? */
	conf.nswap = conf.nproc * 80;
	conf.nswppo = 4096;
	conf.upages = (conf.npage * 70) / 100;
	conf.ialloc = ((conf.npage - conf.upages)/2) * BY2PG;
	conf.copymode = 0;
}

/* declared in port, but only used here */
void
machinit(void)
{
	int n;

	n = m->machno;
	memset(m, 0, sizeof *m);
	m->machno = n;
	m->stack = BOOTSTACK;
}

static void
miscinit(void)
{
	memset(edata, 0, (char*)end - (char*)edata);

	m = &mach0;
	m->machno = 0;
	machinit();
	up = 0;
	active.exiting = 0;
	active.ispanic = 0;
	active.machs = 1;

	//screenputs = xputs;
	consdebug = rdb;
	putasi8(ASI_LSU_CONTROL, 0, Dmmu|Immu|Dcache|Icache);
	savefpregs(&initfp);
}

void
exit(int ispanic)
{
	int once, ms;

	/* ensure orderly shutdown of all cpus */
	lock(&active);
	if(ispanic) 
		active.ispanic = ispanic;
	else if(m->machno == 0 && (active.machs & (1<<m->machno))==0)
		active.ispanic = 0;
	once = active.machs & (1<<m->machno);
	active.ispanic = ispanic;
	active.machs &= ~(1<<m->machno);
	active.exiting = 1;
	unlock(&active);

	if(once)
		print("cpu%d: exiting\n", m->machno);

	/* drain console */
	spllo();
	for(ms = 5*1000; ms > 0; ms -= TK2MS(2)) {
		delay(TK2MS(2));
		if(active.machs == 0 && consactive() == 0)
			break;
	}

	intrdisable();
	xprint("exit.\n");
	reset();
}

void
main(void)
{
	miscinit();
	confinit();
	//archinit();   XXX when we have devarch.c
	xinit();			/* get port:xalloc going */
	mmuinit();
	trapinit();
	z8530init();
	clockinit();

	printinit();		/* get port:devcons going */
	quotefmtinstall();
	print("\n\nPlan 9\n");
 
	procinit0();		/* port:proc */
	initseg();			/* port:segment */
	timersinit();		/* port:portclock */
	links();			/* initiate configured links */
	chandevreset();		/* port:chan */
	pageinit();			/* port:page */
	swapinit();		/* port:swap */
	userinit();			/* build first userland process */
	schedinit();		/* port: start scheduling, doesn't return */

	exit(0);			/* not reached */
}

