#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"

/*
 * XXX we dont ever remove items from the TLBs.  We just let them
 * get replaced with time.  An explicit remove could improve performance.
 *
 * XXX we dont add mappings at kmap() time, we let them fault in.
 * Probing and adding if necessary could improve performance.

 * XXX we need a way to ensure that all the pages
 * that a process' kernel stack can be on will be wired
 * into the TLB!  We cant have a kernel page fault while
 * handling early traps. 
 * Alternately we would need to figure out the fault type,
 * and address very early in l.s before handing off to C to
 * ensure that multiple misses dont overwrite the same
 * global state before the trap handler runs.
 */

#define BIT_ISSET(x, b)	(((x)[(b)/32] & (1 << ((b) & 0x1f))) != 0)
#define BIT_SET(x, b)	(x)[(b)/32] |= (1 << ((b) & 0x1f))
#define BIT_CLR(x, b)	(x)[(b)/32] &= ~(1 << ((b) & 0x1f))
#define TLBADDR(x)		((x)<<3)

#define PGMAPVA(idx)	(MAPSEGBASE + ((idx)<<PGSHIFT))
#define PGMAPIDX(pa)	(((ulong)(pa) - MAPSEGBASE)>>PGSHIFT)


enum {
	KB = 1024,
	NCTX = 256,		// XXX 8192,
	NTLB = 64,
	BY2ICACHELINE = 32,
	BY2DCACHELINE = 16,
	DCACHEMASK = 0x3fff,

	IoMmuControl = 0x2400,
	IoMmuTsb = 0x2408,
	IoMmuFlush = 0x2410,
	SbusControl = 0x2000,
	SbusSlotConfig0 = 0x2020,

	/* IoMmuControl bits */
	IoMmuSz32K = 5<<16,
	IoMmuPg64K = 1<<2,
	IoMmuEn = 1<<0,
};

/* IOMMU PTE bits */
#define IOMMUV	(1LL<<63)
#define IOMMU64K	(1LL<<61)
#define IOMMUSTREAM	(1LL<<60)
#define IOMMUCACHABLE	(1LL<<4)
#define IOMMUW	(1LL<<1)

ulong sbusreg = 0;

/*
 * kernel mappings are either directly mapped using large (4MB) pages
 * or mapped through the kmapalloc table.  Direct mapping is used from
 * VA KZERO through MAPSEGBASE and table mapping is done from
 * MAPSEGBASE to 2^32.
 * Linear mapping is used whenever possible.  Table mapping is only
 * used when linear mapping would violate a page's color or if an
 * arbitrary physical address need be mapped.  Physical mappings are
 * likely to be locked down, and not need replacement.
 * 
 * kmap(pg) returns a mapped kernel virtual address in either of the
 * two regions.  kmapsync(va) installs the mapping in the DTLB when
 * a fault occurs.  kunmap is not necessary (but can improve performance)
 * since stale mappings will eventually be replaced or reused.
 */
static struct {
	Lock;
	ulong count[NCOLOR];
	ulong paddr[NPGMAP];
} kmapalloc = { 0 };

static struct {
	Lock;
	uint used[NCTX/32];
} ctxtab = { 0 };


/*
 * Allocate a context and return it.
 * We have more contexts that processes, so we never
 * need to reclaim contexts.
 */
static int
allocctx(Proc *p)
{
	int i, end;

	/* quickly scan 32 at a time */
	lock(&ctxtab);
	for(i = 0; i < NCTX / 32; i++) {
		if(ctxtab.used[i] != 0xffffffff) {
			end = (i+1) * 32;
			for(i = i * 32; i < end; i++) {
				if(!BIT_ISSET(ctxtab.used, i)) {
					BIT_SET(ctxtab.used, i);
					p->ctx = i;
					unlock(&ctxtab);
					return i;
				}
			}
		}
	}
	unlock(&ctxtab);
	panic("allocctx: too many procs");
	return -1;
}

static void
usectx(Proc *p)
{
	putasi8(ASI_MMU_D, MMU_CTX_PRI, p->ctx);
	putasi8(ASI_MMU_D, MMU_CTX_SEC, p->ctx);
	coherence();
}

static void
freectx(Proc *p)
{
	int s;

	if(p->ctx == 0)
		return;
	s = splhi();
	usectx(p);
	putasi8(ASI_MMU_IDEMAP, DEMAPCTX, 0);
	putasi8(ASI_MMU_DDEMAP, DEMAPCTX, 0);

	lock(&ctxtab);
	BIT_CLR(ctxtab.used, p->ctx);
	unlock(&ctxtab);
	p->ctx = 0;
	usectx(p);
	splx(s);
}

/*
 * Put a page table entry into the kernel's DTLB.
 * Caller is responsible for setting any caching, io or page size fields in phys.
 */
void
putkdtlb(ulong virt, uvlong phys)
{
	/* XXX locking? */
	putasi8(ASI_MMU_D, MMU_TAGACCESS, virt);
	putasi8(ASI_MMU_DDATAIN, 0, phys | PTEVALIDHW | PTEKERNEL | PTEGLOB);
}

/*
 * enable dma on slot0..3,  APC, MACIO/FEPS, SLAVIO -- n = 0..5.
 */
void
sbusenabledma(int n)
{
	/* set ARB_EN[n] */
	put8(sbusreg + SbusControl, get8(sbusreg + SbusControl) | (1<<n));

	/* 32 and 64byte dma, etm, parity, caching */
	put8(sbusreg + SbusSlotConfig0 + n*8, 0xe018);
}

/* init mmu.  Call after xinit. */
void
mmuinit(void)
{
	ulong va;
	int i;

	memset(&ctxtab, 0, sizeof ctxtab);
	
	/* 
	 * at this point we already have a 4MB DTLB and ITLB entry for KZERO
	 * locked in entry 63 thanks to l.s.
	 */
	zsetdev(0);
	for(i = 0; i < NTLB - 1; i++) {
		putasi8(ASI_MMU_IDATA, TLBADDR(i), 0);
		putasi8(ASI_MMU_DDATA, TLBADDR(i), 0);
	}

	/* wire down uart controller for xprint */
	va = (ulong)VA(kmapphys(EIADUART, 1, PTEIO|PTEUNCACHED|PTELOCK));
	zsetdev(va + 4);
#ifdef notneeded
	uvlong endmap, pa;

	/*
	 * map in xalloc pages so they will be available prior to the
	 * trap handler coming up.
	 * XXX is this needed?  The trap handler will come up shortly
	 * and only need to allocate a little bit which will be from low
	 * memory.
	 */
	if(conf.npage1 != conf.base1)
		endmap = PADDR(conf.npage1);
	else
		endmap = PADDR(conf.npage0);
	for(pa = 4*MB; pa < endmap; pa += 4*MB)
		putkdtlb((ulong)KADDR(pa), pa|PTE4MB|PTECACHE|PTEWRITE);
#endif

	/* ctx 0 is reserved for processes without mappings */
	BIT_SET(ctxtab.used, 0);

	/* enforce less contexts than processes. */
	if(conf.nproc >= NCTX)
		panic("mmuinit: nproc too large");

	/* Disable the IOMMU.  DMA addresses are expressed as PADDRs. */
	sbusreg = (ulong)VA(kmapphys(SBUS, 2, PTEIO|PTEUNCACHED|PTELOCK));
	put8(sbusreg + IoMmuControl, 0);
	get8(sbusreg + IoMmuControl);
}

/* 
 * map some region of memory into kernel VM space.
 * Note: if the page isnt wired down, kmapsync() may replace the
 * mapping in the future with the cachable flag enabled.
 */
KMap*
kmapphys(uvlong pa, int pgs, ulong flag)
{
	int idx, i, color;
	ulong va, va0;

	/* find a span of unused VA space */
	ilock(&kmapalloc);
	for(idx = 0; idx < NPGMAP; idx++) {
		if(kmapalloc.paddr[idx] == 0) {
			for(i = 1; i < pgs; i++) {
				if(kmapalloc.paddr[idx+i] != 0)
					break;
			}
			if(i == pgs)
				break;
		}
	}
	if(idx == NPGMAP) {
		xprint("kmapalloc: ");
		for(color = 0; color < NCOLOR; color++)
			xprint("%d ", kmapalloc.count[color]);
		xprint("\n");
		iunlock(&kmapalloc);
		panic("kmapphys: no pages left");
	}
	
	va0 = PGMAPVA(idx);
	va = va0;
	for(i = 0; i < pgs; i++) {
		/* 
		 * we use a dummy pa since these are all locked, and their low
		 * 32bits may read out as zero.  Ugh, more 32-bit hacks.
		 */
		kmapalloc.paddr[idx+i] = 0xffff0000;
		putkdtlb(va, pa|PTELOCK|PTEWRITE|flag);
		pa += BY2PG;
		va += BY2PG;
	}
	iunlock(&kmapalloc);
	return (KMap*)va0;
}

/* port: map a page in to kernel space */
KMap*
kmap(Page *pg)
{
	int color, idx;

	/* find a va with the right color in kmapalloc */
	ilock(&kmapalloc);
	for(idx = pg->color; idx < NPGMAP; idx += NCOLOR) {
		if(kmapalloc.paddr[idx] == 0) {
			/* 
			 * XXX we could add the mapping now if it didnt exist,
			 * instead we just let it fault in.
			 */
			kmapalloc.count[pg->color] ++;
			kmapalloc.paddr[idx] = pg->pa;
			iunlock(&kmapalloc);
			return (KMap*)PGMAPVA(idx);
		}
	}

	xprint("kmapalloc: ");
	for(color = 0; color < NCOLOR; color++)
		xprint("%d ", kmapalloc.count[color]);
	xprint("\n");
	iunlock(&kmapalloc);
	panic("kmapalloc: no pages left!");
	return nil;
}

/* port:  unmap a page from kernel space. */
void
kunmap(KMap *k)
{
	ulong va;
	int color, idx;

	va = (ulong)k;
	if(va < MAPSEGBASE)
		panic("kunmap: not a kmap page!");

	color = getpgcolor(va);
	idx = PGMAPIDX(va);
	ilock(&kmapalloc);
	if(kmapalloc.paddr[idx] == 0)
		panic("kunmap: page not mapped");
	kmapalloc.paddr[idx] = 0;
	kmapalloc.count[color] --;
	putasi8(ASI_MMU_DDEMAP, va, 0);
	iunlock(&kmapalloc);
}

/* replace kernel data TLB entries */
int
kmapsync(ulong va)
{
	ulong pa, idx;

	if(va < KZERO)
		return 0;

	if(va < MAPSEGBASE) {
		/* linear mapping with large pages */
		va &= ~(4*MB - 1);
		putkdtlb(va, PTE4MB | PADDR(va) | PTECACHE | PTEWRITE);
	} else {
		/* table mapped */
		va &= ~(BY2PG-1);
		idx = (va - MAPSEGBASE) >> PGSHIFT;
		pa = kmapalloc.paddr[idx];
		if(!pa)
			return 0;
		putkdtlb(va, PTE8KB | pa | PTECACHE | PTEWRITE);
	}
	return 1;
}

static void
dcflushpg(Page *pg)
{
	ulong i, end, x;

	/* displace cache entries with reads to KZERO */
	i = (pg->va & DCACHEMASK) | KZERO;
	end = i + BY2PG;
	for(; i < end; i += BY2DCACHELINE) {
		x = *(ulong*)i;
		USED(x);
	}
}

static void
icflushpg(Page *pg)
{
	int i;

	for(i = pg->va; i < pg->va + BY2PG; i += BY2ICACHELINE)
		icflush(i);
}

/* 
 * port:  add a MMU mapping for the current process.
 * XXX unfortuantely right now ulong is 32 bits and our addresses are 64
 * bits.  We work around this in the pa bits with software bits.
 *
 * XXX we should probably DEMAP before we add, since there may
 * have previously been a readonly entry that needs to be replaced with
 * a readwrite entry.
 */
void
putmmu(ulong va, ulong smallpa, Page *pg)
{
	uvlong pa;
	char *ctl;
	int s;

	if(!up)
		panic("putmmu !up");
	if(va >= KZERO) {
		pprint("process too large %lux; kiled\n", va);
		pexit("Suicide", 0);
	}

	s = splhi();
	pa = smallpa;
	if(!up->ctx) {
		allocctx(up);
		usectx(up);
	}
	if((pa & PTEUNCACHED) == 0)
		pa |= PTECACHE;
	if(pa & PTEVALID)			/* why wouldnt this be set? */
		pa |= PTEVALIDHW;

	/* 
	 * since there is no distinction between executable and
	 * nonexecutable pages, we put the mapping in both tlbs
	 */
	putasi8(ASI_MMU_IDEMAP, va, 0);
	putasi8(ASI_MMU_DDEMAP, va, 0);
	putasi8(ASI_MMU_I, MMU_TAGACCESS, va | up->ctx);
	putasi8(ASI_MMU_IDATAIN, 0, pa);
	putasi8(ASI_MMU_D, MMU_TAGACCESS, va | up->ctx);
	putasi8(ASI_MMU_DDATAIN, 0, pa);;

	ctl = &pg->cachectl[m->machno]; 
	switch(*ctl) {
	case PG_TXTFLUSH:
		icflushpg(pg);
		*ctl = PG_NOFLUSH;
		break;
	case PG_NEWCOL: /* color changed */
		icflushpg(pg);
		dcflushpg(pg);
		*ctl = PG_NOFLUSH;
		break;
	}
	splx(s);
}


/* port: discard all user memory mappings. */
void
flushmmu(void)
{
	up->newtlb = 1;
	mmuswitch(up);
	coherence();
}

/* port: give back all mmu resources attached to this process. */
void
mmurelease(Proc *p)
{
	freectx(p);		
}

/* 
 * port: A new process is about to run and the port code
 * may have set p->newtlb to request the old mmu data
 * be scrapped.
 */
void
mmuswitch(Proc *p)
{
	if(p->newtlb)
		freectx(p);
	p->newtlb = 0;
	if(!p->ctx)
		allocctx(p);
	usectx(p);
}

/* DEBUG STUFF */
void
dumptlb(int itlb)
{
	int i, off;
	uvlong x, y;

	off = itlb ? 8 : 0;
	xprint("%s:\n", itlb ? "itlb" : "dtlb");
	for(i = 0; i < 64; i++) {
 	 	x = getasi8(ASI_MMU_DTAG - off, i<<3);
 	 	y = getasi8(ASI_MMU_DDATA - off, i<<3);
		if(y & PTEVALIDHW)
 	 		xprint("%d: %x::%x -> %x::%x\n", i, (ulong)(x>>32), (ulong)x ,(ulong)(y>>32), (ulong)y);
	}
}
