/*
 * Driver for the qlogic FAS366 (ncr53c9x variant) SCSI controller
 * using the FEPS/sbus DMA controller (lsi64854).
 * Based heavily on 9.95/ss10/scsi.c.
 *
 * XXX figure out how to go wide.
 * XXX figure out how to access 16 targets.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/sd.h"

/* SCSI DMA registers on FEPS */
typedef struct FepsDma FepsDma;
struct FepsDma {
	ulong	csr;
	ulong	addr;
	ulong	count;
	ulong	status;
};

/* FAS366 hardware registers */
typedef struct FasScsi FasScsi;
struct FasScsi {
	uchar	xfercntlo;		/* 0 */
	uchar		pad1[3];
	uchar	xfercnthi;		/* 4 */
	uchar		pad2[3];
	uchar	fifo;			/* 8 */
	uchar		pad3[3];
	uchar	cmd;			/* c */
	uchar		pad4[3];
	union {				/* 10 */
		uchar	status1;
		uchar	destid;
		int		pad5;
	};
	union {				/* 14 */
		uchar	intr;
		uchar	timeout;
		int		pad6;
	};
	union {				/* 18 */
		uchar	step;
		uchar	syncperiod;
		int		pad7;
	};
	union {				/* 1c */
		uchar	fflags;
		uchar	syncoffset;
		int		pad8;
	};
	uchar	config1;		/* 20 */
	uchar		pad9[3];
	union {				/* 24 */
		uchar	clkconf;
		uchar	status2;
		int		pad10;
	};
	int			pad11;
	uchar	config2;		/* 2c */
	uchar		pad12[3];
	uchar	config3;		/* 30 */
	uchar		pad13[3+4];
	uchar	recntlo;		/* 38 */
	uchar		pad14[3];
	uchar	recnthi;		/* 3c */
};

typedef struct Ctlr Ctlr;
struct Ctlr {
	FepsDma	*dma;		/* dma registers */
	FasScsi	*dev;		/* scsi registers */

	SDev		*sdev;
	SDreq	*req;
	Rendez;
	QLock;
	ulong	xfercnt;		/* start counter of last xfer */
	ushort	id;
	uchar	lastcmd;
};

enum {
	NDev		= 1,
	ScsiId		= 7,
	ClockFreq		= 40 * 1000000,
	TimeoFactor	= 7682,			/* FAS366U section 4.9 */

	/* Dma csr */
	DmaInt		= 1<<0,
	DmaErr		= 1<<1,
	DmaDrain		= 1<<2,
	DmaIntEn		= 1<<4,
	DmaReset		= 1<<7,
	DmaToMem	= 1<<8,
	DmaFromMem	= 0<<8,
	DmaEn		= 1<<9,
	DmaReqPend	= 1<<10,
	DmaWide		= 1<<15,
	DmaDisableDrain = 1<<17,
	DmaBusy		= (1<<2)|(1<<10),

	/* Ctlr command */
	Dma			= 1<<7,
	Nop			= 0x00,
	Flush		= 0x01,
	Reset		= 0x02,
	Busreset		= 0x03,
	Transfer		= 0x10,
	Cmdcomplete	= 0x11,
	Msgaccept		= 0x12,
	Select		= 0x41,

	/* config1 */
	ParityEn		= 1<<4,
	ResetIntDis	= 1<<6,

	/* config2 */
	Hme32		= 1<<7,		/* according to other drivers I've seen */
	FeatureEn		= 1<<4,
	Scsi2En		= 1<<3,

	/* config3 */
	WideEn		= 1<<6,
	FastEn		= 3<<0,

	IntReset		= 1<<7,		/* intr */
	Clk40MHz		= 0,			/* clkconf */
	Active		= 1<<3,		/* status2 */
};

extern SDifc sdfas366ifc;

static Ctlr fas366ctlr[NDev];
static int ndev = 0;

static void
dmawait(FepsDma *dma)
{
	ulong csr;
	int cnt;

	cnt = 0;
	while((csr = dma->csr) && !(csr & DmaErr) && (csr & DmaBusy)) {
		if(cnt++ == 10000) {
			cnt = 0;
			xprint("fas366 long dmawait! %x %x\n", csr, csr & DmaBusy);
		}
		continue;
	}
}

static void
scsireset(FasScsi *dev)
{
	int intr;

	/* dont generate interrupt during reset */
	dev->config1 |= ResetIntDis;
	dev->cmd = Busreset;
	delay(1);
	intr = dev->intr;
	dev->config1 &= ~ResetIntDis;
	USED(intr);
}

/* reset and enable the device */
static int
fas366enable(SDev* sdev)
{
	Ctlr *ctlr;
	FepsDma	*dma;
	FasScsi *dev;
	uchar intr;

	ctlr = sdev->ctlr;
	dev = ctlr->dev;
	dma = ctlr->dma;

	dmawait(dma);
	dma->csr = DmaReset;
	dma->csr = DmaIntEn;

	dev->cmd = Reset;
	delay(1);
	dev->cmd = Dma | Nop;
	dev->cmd = Dma | Nop;
	ctlr->id = (dev->recnthi << 8) | dev->recntlo;
	dev->xfercntlo = 0;
	dev->xfercnthi = 0;
	dev->clkconf = Clk40MHz;
	/* 250msec.  (Clk40MHz specifies 8 periods) */
	dev->timeout = 1 + (250 * (ClockFreq/1000)) / (4 * TimeoFactor);
	dev->syncoffset = 0;
	dev->syncperiod = 4;
	dev->config1 = ParityEn | ScsiId;
	dev->config2 = Hme32 | FeatureEn | Scsi2En;
	/* XXX when I set wide, the data is interleaved with NUL bytes 
	 * which is probably to be expected but I'm still a scsi dunce.
	 */
	dev->config3 = FastEn; /* WideEn | FastEn; */
	dev->recntlo = 0;
	dev->recnthi = 0;
	dev->cmd = Dma | Nop;
	intr = dev->intr;
	USED(intr);
	scsireset(dev);

	print("#S fas366 %x\n", ctlr->id);
	return 1;
}

static void
warn(char *msg, ulong status, ulong step, ulong intr, ulong csr)
{
	xprint("sdfas366intr: %s status=%x, step=%x, intr=%x, csr=%x\n", 
		msg, status, step, intr, csr);
}

/* start a DMA transfer or complete transaction */
static void
startxfer(Ctlr *ctlr, int phase, uchar status, uchar step, uchar intr, ulong csr)
{
	FepsDma *dma;
	FasScsi *dev;
	SDreq *r;
	int frommem;

	dma = ctlr->dma;
	dev = ctlr->dev;
	r = ctlr->req;

	ctlr->xfercnt = r->dlen - r->rlen;
	if(ctlr->xfercnt > 64 * 1024)
		ctlr->xfercnt = 64 * 1024;

	if(ctlr->xfercnt > 0	
	&& ((phase == 0 && r->write)
	|| (phase == 1 && !r->write))) {
		frommem = (phase == 0);
		/* and continue below to start the transfer */
	} else if(phase == 3) {
		dev->cmd = ctlr->lastcmd = Cmdcomplete;
		return;
	} else {
		warn("weird phase after xfer", status, step, intr, csr);
		dev->cmd = ctlr->lastcmd = Cmdcomplete;
		return;
	}

	dmawait(dma);
	dma->csr = DmaReset;
	dma->csr = DmaIntEn;
	dma->addr = PADDR((char*)r->data + r->rlen);
	dma->count = ctlr->xfercnt;
	dma->csr = DmaDisableDrain | /* DmaWide |*/ DmaEn | DmaIntEn | (frommem ? DmaFromMem : DmaToMem);

	dev->xfercntlo = ctlr->xfercnt;
	dev->xfercnthi = ctlr->xfercnt >> 8;
	ctlr->lastcmd = dev->cmd = Dma | Transfer;
}

static void
fas366intr0(Ctlr *ctlr)
{
	FepsDma	*dma;
	FasScsi *dev;
	SDreq *r;
	ulong csr, reset, cnt;
	uchar status, step, intr, phase, msg;
	int i;

	dev = ctlr->dev;
	dma = ctlr->dma;
	r = ctlr->req;

	csr = dma->csr;
	if(csr & DmaErr) {
		print("fas366 dma error!");
		reset = 1;
		goto complete;
	}
	if(!(csr & DmaInt))
		return;

	status = dev->status1;
	step = dev->step;
	intr = dev->intr;		/* read clears status,step,intr */

	//xprint("intr csr %x status %x step %x intr %x last cmd %x fflags %x status2 %x\n", csr, status, step, intr, ctlr->lastcmd, dev->fflags, dev->status2);

	step = step & 7;
	phase = status & 7;
	if(!r) {
		warn("spurious", 0, step, intr, csr);
		return;
	}

	reset = 0;
	if(intr & IntReset) {
		xprint("SCSI reset\n");
		dev->cmd = Nop;
		r->status = SDeio;
	} else {
		/* XXX what return values are right here? */
		switch(ctlr->lastcmd & ~Dma) {
		case Select:
			switch((step << 8) | intr) {
			case 0x020:		/* select timed out */
				r->status = SDtimeout;
				break;
			case 0x218:		/* arb, select but no cmd*/
				warn("no cmd phase", r->status, step, intr, csr);
				reset = 1;
				break;
			case 0x318:		/* cmd phase ended prematurely */
				/* XXX can still be recovered sometimes? */
				warn("short cmd phase", r->status, step, intr, csr);
				reset = 1;
				break;
			case 0x418:		/* arb,select,cmd done */
				startxfer(ctlr, phase, status, step, intr, csr);
				return;
			default :
				warn("illegal hardware state", r->status, step, intr, csr);
				reset = 1;
				break;
			}
			break;

		case Transfer:
			/* update rlen based on what we tried to send and whats left */

			/* XXX do we have to pop out an extra item from the fifo? */
			if(dev->status2 & 0x60) 
				panic("fifo not drained when handling xfer");

			/* XXX xfercnt of zero can mean two things! */
			cnt = (dev->xfercnthi<<8) | dev->xfercntlo;
			r->rlen += ctlr->xfercnt - cnt;
			startxfer(ctlr, phase, status, step, intr, csr);
			return;
			
		case Cmdcomplete:
			r->status = dev->fifo;
			msg = dev->fifo;
			USED(msg);
			dev->cmd = ctlr->lastcmd = Msgaccept;
			return;

		case Msgaccept:
			if(phase == 2) {
				for(i = 0; i < r->clen; i++)
					dev->fifo = r->cmd[i];
				dev->cmd = ctlr->lastcmd = Transfer;
				return;
			}
			break;

		default:
			warn("unexpected state", status, step, intr, csr);
			reset = 1;
			break;
		}
	}

	/* complete the request */
complete:
	if(reset) {
		print("SCSI RESET\n");
		if(r)
			r->status = SDeio;
		fas366enable(ctlr->sdev);
	}
	ctlr->req = 0;
	dma->csr = DmaIntEn;
	if(r)
		wakeup(ctlr);
}

void
fas366intr(void)
{
	Ctlr *ctlr;
	int i;

	for(i = 0; i < ndev; i++) {
		ctlr = &fas366ctlr[i];
		fas366intr0(ctlr);
	}
}

static int
fas366done(void *arg)
{
	Ctlr *ctlr;

	ctlr = (Ctlr*)arg;
	return (ctlr->req == 0);	
}

static int
fas366rio0(SDreq* r)
{
	Ctlr *ctlr;
	FasScsi *dev;
	FepsDma *dma;
	int i;

	ctlr = r->unit->dev->ctlr;
	dev = ctlr->dev;
	dma = ctlr->dma;

	/* XXX
	xprint("req: unit %d %s cmdlen %d len %d -- ", r->unit->subno, 
		r->write ? "write" : "read", r->clen, r->dlen);
	for(i = 0; i < r->clen; i++) xprint("%x ", r->cmd[i]);
	xprint("\n");
	/**/

	if(r->clen != 6 && r->clen != 10 && r->clen != 12)
		return SDeio;
	if(r->unit->subno == ScsiId)
		return SDeio;

	/* prepare to dma the data */
	coherence();

	/* clear fifo and load it up with our cmd */
	qlock(ctlr);
	dev->cmd = Flush;
	for(i = 0; i < r->clen; i++) {
		dev->fifo = r->cmd[i];
		dev->fifo = 0;			/* fas366 weirdness */
	}

	/* fire it off and wait for it to complete */
	dma->csr = DmaIntEn;
	r->status = 0;
	r->rlen = 0;
	ctlr->req = r;
	dev->destid = (r->unit->subno & 7) | 0x10;
	dev->cmd = ctlr->lastcmd = Select;
	while(waserror())
		continue;
	sleep(ctlr, fas366done, ctlr);
	poperror();
	qunlock(ctlr);

	/* dma is completed */
	coherence();

	// XXX xprint("status: %x  len %x\n", r->status, r->rlen);
	return r->status;
}

static int
fas366rio(SDreq* r)
{
	if(fas366rio0(r) == SDcheck && !(r->flags & SDnosense)) {
		r->write = 0;
		memset(r->cmd, 0, sizeof(r->cmd));
		r->cmd[0] = 0x03;
		r->cmd[1] = r->lun<<5;
		r->cmd[4] = sizeof(r->sense)-1;
		r->clen = 6;
		r->data = r->sense;
		r->dlen = sizeof(r->sense)-1;
		if(fas366rio0(r) == SDok) {
			r->flags |= SDvalidsense;
			r->status = SDcheck;
		}
	}
	return r->status;
}


static SDev*
fas366id(SDev* sdev)
{
	return scsiid(sdev, &sdfas366ifc);
}

/* 
 * Get the FEPS/FAS366 controller on the sbus.
 */
static SDev*
fas366sbus(void)
{
	Ctlr *ctlr;
	SDev *sdev;
	KMap *k;

	if(ndev >= NDev)
		panic("too many fas366 devices");
	ctlr = &fas366ctlr[ndev++];
	k = kmapphys(SCSI, 1, PTEIO|PTEUNCACHED|PTELOCK);
	ctlr->dev = (void*)VA(k);
	k = kmapphys(SCSIDMA, 1, PTEIO|PTEUNCACHED|PTELOCK);
	ctlr->dma = (void*)VA(k);

	sdev = malloc(sizeof *sdev);
	if(!sdev)
		panic("Can't allocate SCSI device");
	sdev->ifc = &sdfas366ifc;
	sdev->ctlr = ctlr;
	ctlr->sdev = sdev;
	sdev->nunit = 8;			/* XXX 16 targets? */

	ctlr->dma->csr = DmaReset;
	delay(1);
	ctlr->dma->csr = 0;
	sbusenabledma(SbusFeps);
	sbusenableintr(SbusIntrMapScsi);
	return sdev;
}

static SDev*
fas366pnp(void)
{
	SDev *head;

	head = fas366sbus();
	return head;
}

SDifc sdfas366ifc = {
	"fas366",			/* name */

	fas366pnp,		/* pnp */
	nil,				/* legacy */
	fas366id,			/* id */
	fas366enable,		/* enable */
	nil,				/* disable */

	scsiverify,			/* verify */
	scsionline,		/* online */
	fas366rio,			/* rio */
	nil,				/* rctl */
	nil,				/* wctl */

	scsibio,			/* bio */
	nil,				/* probe */
	nil,				/* clear */
	nil,				/* stat */
};
