#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "ureg.h"

#include "init.h"

enum {
	ARGSIZE = 128,	/* max space for init0 arguments */ 
};

static char *av0[2] = { "/sparc64/9ultra2", 0};

/* 
 * port: Give enough context in the ureg to produce
 * a kernel stack for a sleeping process.
 */
void
setkernur(Ureg *ur, Proc *p)
{
	/* XXX verify that this is correct. */
	ur->pc = p->sched.pc;
	ur->sp = p->sched.sp;
	ur->r15 = (ulong)sched;
}

/* port: Set registers from user provided registers. */
void
setregisters(Ureg *ur, char *pureg, char *uva, int n)
{
	Ureg *pur;
	ulong pstate;

	pstate = ur->pstate;
	memmove(pureg, uva, n);
	pur = (Ureg*)pureg;
	pur->pstate = (pur->pstate & ~0xffff) | (pstate & 0xffff);
}

/* port: set registers for exec to return to entry */
long
execregs(ulong entry, ulong ssize, ulong nargs)
{
	Ureg *ur;
	ulong *sp;

	sp = (ulong*)(USTKTOP - ssize);
	*--sp = nargs;

	ur = (Ureg*)up->dbgreg;
	ur->usp = (ulong)sp;
	ur->pc = entry - 4;	/* syscall advances pc by four */
	ur->pstate &= ~PSPEF;
	up->fpstate = FPinit;
	return USTKTOP-BY2WD;	/* address of user-level clock */
}

static void
_dumpstack(Ureg *ur)
{
	ulong l, v, i, estack;
	extern ulong etext;

	iprint("dumpstack\n");
	iprint("ktrace /kernel/path %.8lux %.8lux\n", ur->pc, ur->sp);

	/* figure out which stack region we're in */
	if(up && (ulong)&l >= (ulong)up->kstack &&
			(ulong)&l <= (ulong)up->kstack+KSTACK)
		estack = (ulong)up->kstack+KSTACK;
	else if((ulong)&l >= (ulong)m->stack &&
			(ulong)&l <= (ulong)m->stack+KSTACK)
		estack = (ulong)up->kstack+KSTACK;
	else
		return;
	print("estackx %.8lux\n", estack);

	i = 0;
	for(l = (ulong)&l; l < estack; l+=4) {
		v = *(ulong*)l;
		if((KTZERO < v && v < (ulong)&etext) || estack-l<256){
			iprint("%.8lux=%.8lux ", l, v);
			i++;
		}
		if(i == 4){
			i = 0;
			iprint("\n");
		}
	}
	if(i)
		iprint("\n");
}

/* port */
void
dumpstack(void)
{
	callwithureg(_dumpstack);
}

/* port: Return the userland PC for fault reporting and process dumping */
ulong
dbgpc(Proc *p)
{
	Ureg *ur;

	ur = p->dbgreg;
	if(ur)
		return ur->pc;
	return 0;
}

/* port */
ulong
userpc(void)
{
	return dbgpc(up);
}

/* port */
void
procsetup(Proc *p)
{
	p->fpstate = FPinit;
}

/* port */
void
procrestore(Proc *p)
{
	uvlong t;

	if(p->kp)
		return;
	cycles(&t);
	p->pcycles -= t;
}

/* port */
void
procsave(Proc *p)
{
	uvlong t;

	cycles(&t);
	p->pcycles += t;
	fppause();

	/* XXX see comment in pc/main.c procsave when we do SMP */
}

/* port: call function with a copy of registers. */
void
callwithureg(void(*fn)(Ureg*))
{
	Ureg ur;

	memset(&ur, 0, sizeof ur);
	ur.sp = (ulong)&fn - BY2WD;
	// XXX ur.r2 = setSB;
	ur.r5 = (ulong)up;
	ur.r6 = (ulong)m;
	ur.r7 = (ulong)fn;
	ur.r15 = getcallerpc(&fn);
	ur.pc = (ulong)callwithureg;
	fn(&ur);
}

static void
linkproc(void)
{
	spllo();
	up->kpfun(up->kparg);
	pexit("kproc exiting", 0);
}

/* port: set kernel process to return to func with args. */
void
kprocchild(Proc *p, void (*func)(void*), void *arg)
{
	/* gotolabel adds 8 to the pc */
	p->sched.pc = (ulong)linkproc - 8;
	p->sched.sp = (ulong)p->kstack + KSTACK;

	p->kpfun = func;
	p->kparg = arg;
}

/* kernel portion of first process */
static void
init0(void)
{
	char buf[2*KNAMELEN];
	int i;

	up->nerrlab = 0;
	spllo();

	/*
	 * These are o.k. because rootinit is null.
	 * Then early kproc's will have a root and dot.
	 */
	up->slash = namec("#/", Atodir, 0, 0);
	cnameclose(up->slash->name);
	up->slash->name = newcname("/");
	up->dot = cclone(up->slash);

	kproc("alarm", alarmkproc, 0);
	chandevinit();
	if(!waserror()){
		snprint(buf, sizeof buf, "sparc64 %s", conffile);
		ksetenv("terminal", buf, 0);
		ksetenv("cputype", "sparc64", 0);
		if(cpuserver)
			ksetenv("service", "cpu", 0);
		else
			ksetenv("service", "terminal", 0);
		for(i = 0; i < conf.nprops; i++){
			if(conf.propname[i][0] != '*')
				ksetenv(conf.propname[i], conf.propvalue[i], 0);
			ksetenv(conf.propname[i], conf.propvalue[i], 1);
		}
		poperror();
	}

	/* jump out to userland! */
	touser(USTKTOP - MAXSYSARG * BY2WD - ARGSIZE);
}

static void
bootargs(ulong base)
{
	char **pav, *p;
	ulong sp;
	int i, ac, l;

	sp = base + BY2PG - MAXSYSARG * BY2WD - ARGSIZE;

	for(l = 0, ac = 0; av0[ac]; ac++)
		l += strlen(av0[ac]) + 1;
	ac ++;
	if(l + ac * BY2WD > ARGSIZE)
		panic("bootargs");

	/* 
	 * copy the argvec and strings out to the stack
	 * while relocating the pointers.
	 */
	pav = (char**)sp;
	p = (char*)(pav + ac);
	for(i = 0; av0[i]; i++) {
		pav[i] = p - base + (USTKTOP - BY2PG);
		strcpy(p, av0[i]);
		p += strlen(p);
		*p++ = 0;
	}
	pav[i] = 0;
}

/* build the first userland process */
void
userinit(void)
{
	Proc *p;
	Segment *s;
	KMap *k;
	Page *pg;

	p = newproc();
	p->pgrp = newpgrp();
	p->egrp = smalloc(sizeof *p->egrp);
	p->egrp->ref = 1;
	p->fgrp = dupfgrp(nil);
	p->rgrp = newrgrp();
	p->procmode = 0640;

	kstrdup(&eve, "");
	kstrdup(&p->text, "*init*");
	kstrdup(&p->user, eve);

	procsetup(p);

	/* kernel stack */
	p->sched.pc = (ulong)init0 - 8;	/* we jump to pc+8 */
	p->sched.sp = (ulong)p->kstack + KSTACK - MAXSYSARG*BY2WD;
	p->sched.sp &= ~7;

	/* user stack with boot arguments */
	s = newseg(SG_STACK, USTKTOP - USTKSIZE, USTKSIZE / BY2PG);
	p->seg[SSEG] = s;
	pg = newpage(1, 0, USTKTOP - BY2PG);
	segpage(s, pg);
	k = kmap(pg);
	bootargs(VA(k));
	kunmap(k);

	/* Text from initcode */
	s = newseg(SG_TEXT, UTZERO, 1);
	s->flushme ++;
	p->seg[TSEG] = s;
	pg = newpage(1, 0, UTZERO);
	memset(pg->cachectl, PG_TXTFLUSH, sizeof pg->cachectl);
	segpage(s, pg);
	k = kmap(s->map[0]->pages[0]);
	memmove((uchar*)VA(k), initcode, sizeof initcode);
	kunmap(k);

	ready(p);
}
