#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "ureg.h"
#include "../port/error.h"
#include "../port/systab.h"
#include "io.h"

enum { 
	KB = 1024,
	NINT = 5,
	NDBG = 64,

	/* interrupt levels */
	IntrDefer = 6,
	IntrTick = 14,
};

static ulong trapdbg[NDBG] = {0};
static int dbgpos = 0;
static ulong intstack[NINT] = {0};
static int intcnt = 0;


static char *fptrapname[]= { 
	"none",
	"IEEE 754 exception",
	"unfinished FP op",
	"unimplemented FP op",
	"sequence error",
	"hardware error",
	"invalid FP register",
	"reserved",
};

/* DEBUG STUFF */
static void
dumpureg(Ureg *ur)
{
	int *reg = (int*)ur;
	int i;

	xprint("ureg %x:\n", ur);
	for(i = 0; i < 32; i+=4) {
		xprint("%d: %x %x %x %x\n", i, reg[0], reg[1], reg[2], reg[3]);
		reg += 4;
	}
	xprint("pc %x npc %x tt %x pstate %x\n", ur->pc, ur->npc, ur->tt, ur->pstate);
}


/* write an interrupt vector to target */
static void
putvec(ulong vec, void (*targ)(void))
{
	ulong *x, a;

	/* XXX we could definitely be more efficient here by
	 * using the delay slot and putting more code in the table.
	 */
	x = (ulong*)TRAPS + vec*8;
	a = ((ulong*)targ - &x[1]) & ~0xc0000000;
	x[0] = 0x93902001;		/* MOV $1, RCWP */
	x[1] = 0x40000000 | a;	/* CALL targ(SB) */
	x[2] = 0x80000000;		/* nop */
}

/* xinit and mmuinit must have previously been called */
void
trapinit(void)
{
	ulong i;
	int map;

	/* Disable sbus interrupt vectors. */
	for(map = SbusIntrMapLo; map < SbusIntrMapHi; map += 8) {
		i = get8(sbusreg+map) & 0x7ff;
		if(i)
			put8(sbusreg+map, i);
	}

	/* all vectors to traplink except 0x100 to syslink. */
	for(i=0; i<0x180; i++) {
		putvec(i, traplink);
		putvec(i+0x200, traplink);
	}
	putvec(0x100, syslink);
	coherence();
	puttba(TRAPS);
	intrenable();
}

/* last stop before user space */
static void
kexit(Ureg* ur)
{
	if(up->fpstate == FPactive)
		ur->pstate |= PSPEF;
	else
		ur->pstate &= ~PSPEF;
	if(getpil() != 0) panic("kexit PIL %lux", getpil()); //XXX
}

/* enable sbus interrupts */
void
sbusenableintr(ulong mapoff)
{
	ulong intr;

	intr = get8(sbusreg+mapoff) & 0x7ff;
	put8(sbusreg + mapoff, intr);
	put8(sbusreg + mapoff + (SbusIntrClrLo - SbusIntrMapLo), 0);
	put8(sbusreg + mapoff, intr | 0x80000000LL);
}

static char*
excname(ulong tt, char *excbuf)
{
	if(tt >= 0x41 && tt <= 0x4f) {
		sprint(excbuf, "trap: interrupt level %lud", tt - 0x40);
		return excbuf;
	}
	if(tt >= 0x80 && tt <= 0xbf)
		return "spill";
	if(tt >= 0xc0 && tt <= 0xff)
		return "fill";
	if(tt >= 0x102 && tt <= 0x17f) {
		sprint(excbuf, "trap: trap %lud", tt - 0x100);
		return excbuf;
	}
	if(tt == 0x21 || tt == 0x22) {
		if(up) {
			sprint(excbuf, "fp: %s fsr=0x%lux",
				fptrapname[(up->fpsave.fsr>>14)&7], up->fpsave.fsr);
			return excbuf;
		}
		return "trap: fp exception";
	}

	switch(tt) {
	case 0x08: return "trap: instruction access exception";
	case 0x09: return "trap: instruction access mmu miss";
	case 0x0a: return "trap: instruction access error";
	case 0x10: return "trap: illegal instruction";
	case 0x11: return "trap: privileged opcode";
	case 0x20: return "trap: fp disabled";
	case 0x23: return "trap: tag overflow";
	case 0x24: return "trap: clean window";
	case 0x28: return "trap: division by zero";
	case 0x29: return "trap: processor error";
	case 0x30: return "trap: data access exception";
	case 0x31: return "trap: data access mmu miss";
	case 0x32: return "trap: data access error";
	case 0x33: return "trap: data access protection";
	case 0x37: return "trap: privileged action";
	case 0x40: return "trap: async data error";

	case 0x34: case 0x35: case 0x36:
	case 0x38: case 0x39:
		return "trap: data alignment";

	case 0x12: case 0x13:
		return "trap: unimplemented instruction";

	case 0x100: return "syscall";
	case 0x101: return "breakpoint";

	default:
		sprint(excbuf, "trap: unknown trap %lud", tt);
		return excbuf;

	/* ultrasparc specific */
	case 0x64: return "trap: fast instruction MMU miss";
	case 0x68: return "trap: fast data MMU miss";
	case 0x6c: return "trap: fast data protection";
	}
}

static void
intr(Ureg *ur, int level)
{
	ulong intno, mask;

	mask = 1<<level;
	intrclear(mask);
	switch(level) {
	case IntrTick:		/* %tick compare */
		intrclear(1);	/* can also get here on softint 0 */
		clockintr(ur);
		break;
	case IntrDefer:		/* scheduled soft int */
		for(;;) {
			intrdisable();
			/* XXX debug */
			if(intcnt < 0 || intcnt > NINT)
				panic("intstack blown");

			intno = -1;
			if(intcnt > 0)
				intno = intstack[--intcnt];
			intrenable();
			if(intno == -1)
				break;

			switch(intno) {
			case SbusIntrSerial:
				z8530intr();
				put8(sbusreg + SbusIntrClrSerial, 0);
				break;
			case SbusIntrHme:
				hmeintr();
				put8(sbusreg + SbusIntrClrHme, 0);
				break;
			case SbusIntrScsi:
				fas366intr();
				put8(sbusreg + SbusIntrClrScsi, 0);
				break;
			case SbusIntrTimer1:
				/* ignore leftover timer interrupts from ofw */
				break;
			default:
				xprint("spurious interrupt number %x\n", intno);
			}
		}
		break;
	default:
		xprint("spurious interrupt level %d\n", level);
	}

	/* check preempt for anything but clock */
	if(level != IntrTick)
		preempted();
}


static void
mmumiss(ulong va, int read, Ureg *ur, int user)
{
	ulong addr;
	int x;
	char buf[ERRMAX], *name;

	addr = va & ~(BY2PG-1);
	if(!user && kmapsync(addr))
		return;

	if(!up)
		panic("!up - kernel fault pc=%lux addr=%lux ureg=%lux", ur->pc, va, (ulong)ur);

	/* bill this to system time. */
	x = up->insyscall;
	up->insyscall = 1;

//xprint("mmumiss pc %x addr %x va %x, read %d\n", ur->pc, addr, va, read);
	if(fault(addr, read) < 0) {
//xprint("mmumiss pc %x addr %x va %x, read %d\n", ur->pc, addr, va, read);
		if(!user)
			panic("kernel fault pc=%lux addr=%lux ureg=%lux", ur->pc, va, (ulong)ur);
		name = read ? "read" : "write";
		sprint(buf, "sys: trap: fault %s addr=0x%lux", name, va);
		postnote(up, 1, buf, NDebug);
	}
	up->insyscall = x;
}


static int
fpenable(void)
{
	if(up->fpstate == FPinit)
		restorefpregs(&initfp);
	else if(up->fpstate == FPinactive)
		restorefpregs(&up->fpsave);
	else
		return 0;
	up->fpstate = FPactive;
	return 1;
}

void
fppause(void)
{
	if(up->fpstate == FPactive) {
		savefpregs(&up->fpsave);
		up->fpstate = FPinactive;
	}
}

/* 
 * prototype in portfns.h but only used here 
 * Called at splhi().
 */
int
notify(Ureg *ur)
{
	int l, sz;
	ulong sp, s;
	Note *n;

	if(up->procctl)
		procctl(up);
	if(up->nnote == 0)
		return 0;

	fppause();
	s = spllo();
	qlock(&up->debug);
	up->notepending = 0;
	n = &up->note[0];
	if(strncmp(n->msg, "sys:", 4) == 0) {
		l = strlen(n->msg);
		if(l > ERRMAX-15)	/* " pc=0x12345678\0" */
			l = ERRMAX-15;
		sprint(n->msg+l, " pc=0x%lux", ur->pc);
	}

	if(n->flag != NUser && (up->notified || up->notify==0)) {
		if(n->flag == NDebug)
			pprint("suicide: %s\n", n->msg);
		qunlock(&up->debug);
		pexit(n->msg, n->flag!=NDebug);
	}

	if(up->notified) {
		qunlock(&up->debug);
		splx(s);
		return 0;
	}
		
	if(!up->notify) {
		qunlock(&up->debug);
		pexit(n->msg, n->flag!=NDebug);
	}

	up->fpstate |= FPillegal;
	sp = ur->usp & ~(BY2V-1);
	sz = ERRMAX + 4*BY2WD + sizeof *ur;
	if(!okaddr((ulong)up->notify, BY2WD, 0)
	|| !okaddr(sp - sz, sz, 1)) {
		pprint("suicide: bad address or sp in notify\n");
		qunlock(&up->debug);
		pexit("Suicide", 0);
	}

	sp -= sizeof *ur;
	*(Ureg*)sp = *ur;						/* push ureg */
	up->ureg = (void*)sp;
	sp -= BY2WD;
	*(Ureg**)sp = up->ureg;					/* push &ureg */
	sp -= ERRMAX;
	memmove((char*)sp, up->note[0].msg, ERRMAX);	/* push msg */
	sp -= 3*BY2WD;
	*(ulong*)(sp+2*BY2WD) = sp+3*BY2WD;		/* arg 2 is &msg */
	*(ulong*)(sp+1*BY2WD) = (ulong)up->ureg;	/* arg 1 0(FP) is &ureg* */
	*(ulong*)(sp+0*BY2WD) = 0;				/* arg 0 is return addr */
	ur->r7 = (ulong)up->ureg;				/* r7 caches arg1 */
	ur->r15 = 0;							/* r15 caches return addr */
	ur->usp = sp;
	ur->pc = (ulong)up->notify;
	ur->npc = ur->pc + 4;
	up->notified = 1;
	up->nnote--;
	memmove(&up->lastnote, &up->note[0], sizeof(Note));
	memmove(&up->note[0], &up->note[1], up->nnote*sizeof(Note));

	qunlock(&up->debug);
	splx(s);
	return 1;
}

/*
 * Return user to state before notify().
 */
static void
noted(Ureg *ur, ulong arg0)
{
	Ureg *nur;
	ulong oureg, sp;

	qlock(&up->debug);
	if(arg0 != NRSTR && !up->notified) {
		qunlock(&up->debug);
		pprint("call to noted() when not notified\n");
		pexit("Suicide", 0);
	}
	up->notified = 0;

	nur = up->ureg;	/* pointer to user returned Ureg struct */
	up->fpstate &= ~FPillegal;

	oureg = (ulong)nur;
	if((oureg & (BY2WD-1))
	|| !okaddr((ulong)oureg-BY2WD, BY2WD+sizeof(Ureg), 0)){
		pprint("bad ureg in noted\n");
		qunlock(&up->debug);
		pexit("Suicide", 0);
	}

	/* use the user provided registers (but safely) */
	setregisters(ur, (char*)ur, (char*)nur, sizeof *nur);
	switch(arg0) {
	case NCONT:
	case NRSTR:
		if(!okaddr(nur->pc, BY2WD, 0) || !okaddr(nur->usp, BY2WD, 0)){
			pprint("suicide: trap in noted\n");
			qunlock(&up->debug);
			pexit("Suicide", 0);
		}
		up->ureg = (Ureg*)(*(ulong*)(oureg-BY2WD));
		qunlock(&up->debug);
		break;

	case NSAVE:
		if(!okaddr(nur->pc, BY2WD, 0) || !okaddr(nur->usp, BY2WD, 0)){
			pprint("suicide: trap in noted\n");
			qunlock(&up->debug);
			pexit("Suicide", 0);
		}
		qunlock(&up->debug);
		sp = oureg-4*BY2WD-ERRMAX;
		splhi();
		ur->sp = sp;
		((ulong*)sp)[1] = oureg;	/* arg 1 0(FP) is ureg* */
		((ulong*)sp)[0] = 0;		/* arg 0 is pc */
		ur->r7 = oureg;		/* arg 1 is ureg* */
		break;

	default:
		pprint("unknown noted arg 0x%lux\n", arg0);
		up->lastnote.flag = NDebug;
		/* fall through */
		
	case NDFLT:
		if(up->lastnote.flag == NDebug)
			pprint("suicide: %s\n", up->lastnote.msg);
		qunlock(&up->debug);
		pexit(up->lastnote.msg, up->lastnote.flag != NDebug);
	}
}

/* handle all traps other than syscall */
void
trap(Ureg *ur)
{
	int user, err, pil;
	ulong ttype, intlevel, intno, va, fsr;
	char excbuf[64], buf[64];

	pil = getpil();
	ttype = ur->tt;
	user = !(ur->pstate & PSPRIV);
	if(user)
		up->dbgreg = ur;

	trapdbg[dbgpos++] = ttype;
	trapdbg[dbgpos++] = user;
	trapdbg[dbgpos++] = ur->pc;
	trapdbg[dbgpos++] = pil;
	dbgpos &= (NDBG-1);

	/* Translate intr 0x60 into a softint.  Move this to l.s */
	if(ttype == 0x60) {
		intno = getasi8(ASI_UDB, UDB_DATA0);
		putasi8(ASI_INTR_RECEIVE, 0, 0);
		if(intcnt >= NINT)
			panic("trap: too many interrupts");
		intstack[intcnt++] = intno;
		intrraise(1<<IntrDefer);
		splx(pil);
		return;
	}

	/* XXX consider making intr a seperate trap entry point */
	err = 0;
	intlevel = ttype - 0x40;
	if(intlevel < 16 && intlevel) {
		splx(intlevel);
		intrenable();
		intr(ur, intlevel);
	} else {
		/* 
		 * We should get the fault address in l.s 
		 * to avoid this complex intrenable() handling.
		 */
		switch(ttype) {
		case 0x08:			/* instr. access */
		case 0x09:			/* instr. access miss */
		case 0x64:			/* ultrasparc fast instruction miss */	
			/* instruction fault at %pc */
			intrenable();
			mmumiss(ur->pc, 1, ur, user);
			break;
	
		case 0x30:			/* data access */
		case 0x33:			/* data access protection */
		case 0x6c:			/* ultrasparc fast data protection */
			/* data fault in FAR */
			va = getasi8(ASI_MMU_D, MMU_FAULTADDR);
			fsr = getasi8(ASI_MMU_D, MMU_FAULTSTATUS);
			intrenable();
			mmumiss(va, !(fsr & FSR_W), ur, user);
			break;

		case 0x31:			/* data access miss */
		case 0x68:			/* ultrasparc fast data miss */
			/* data fault in TAGACCESS */
			/* XXX throws away some info about address and read/write! -- can we do better? */
			va = getasi8(ASI_MMU_D, MMU_TAGACCESS) & ~(BY2PG-1);
			intrenable();
			mmumiss(va, 1, ur, user);
			break;

		case 0x40:		/* async data error */
			intrenable();
			panic("kernel trap: asynchronous data error.  AFSR %llux, AFAR %llux",
				getasi8(ASI_AFSR, 0), getasi8(ASI_AFAR, 0));
			break;
	
		case 0x20:			/* floating point disabled */
			intrenable();	
			if(err && up && (up->fpstate & FPillegal))
				postnote(up, 1, "sys: floating point in note handler", NDebug);
			else
				err = !user || !fpenable();
			break;
		case 0x21:			/* fp exception ieee */
		case 0x22:			/* fp exception other */
			intrenable();
			fppause();
			err = 1;
			break;
		default:
			intrenable();
			err = 1;
			break;
		}
	}
	if(err) {
		if(user) {
			sprint(buf, "sys: %s", excname(ttype, excbuf));
			postnote(up, 1, buf, NDebug);
		} else {
			dumpureg(ur);
			panic("kernel trap: %s pc=0x%lux\n", excname(ttype, excbuf), ur->pc);
		}
	}

	/* if we delayed scheduling, do it now */
	splhi();
	if(up && up->delaysched) {
		sched();
		splhi();
	}

	if(user) {
		if(up->procctl || up->nnote)
			notify(ur);
		splx(pil);
		kexit(ur);
	} else
		splx(pil);
}


/* port */
long
syscall(Ureg *ur)
{
	int i, s;
	char *e;
	long ret;
	ulong sp, scallnr;

	splhi();
	if(ur->pstate & PSPRIV)
		panic("kernel syscall from %lux", ur->pc);

	/* XXX cycles(&up->kentry); ? */

	m->syscall++;
	up->insyscall = 1;
	up->pc = ur->pc;
	up->dbgreg = ur;

	if(up->procctl == Proc_tracesyscall) {
		up->procctl = Proc_stopme;
		procctl(up);
	}

	scallnr = up->scallnr = ur->r7;
	up->scallnr = scallnr;
	if(scallnr == RFORK)
		fppause();
	spllo();

	sp = ur->usp;
	up->nerrlab = 0;
	ret = -1;
	if(!waserror()) {
		if(scallnr >= nsyscall || systab[scallnr] == nil) {
			pprint("bad sys call %d pc %lux\n", scallnr, ur->pc);
			postnote(up, 1, "sys: bad sys call", NDebug);
			error(Ebadarg);
		}

		if(sp & (BY2WD-1)) {
			pprint("odd sp in sys call pc %lux sp %lux\n",
				ur->pc, ur->sp);
			postnote(up, 1, "sys: odd stack", NDebug);
			error(Ebadarg);
		}

		if(sp < (USTKTOP-BY2PG) || sp > (USTKTOP - sizeof(Sargs)))
			validaddr(sp, sizeof(Sargs), 0);

		up->s = *((Sargs*)(sp + 1*BY2WD));
//xprint("%s:%d pc %x, %s: %x %x %x %x\n", up->text, up->pid, up->pc, sysctab[scallnr], up->s.args[0], up->s.args[1], up->s.args[2], up->s.args[3]);
		up->psstate = sysctab[scallnr];
		ret = systab[scallnr](up->s.args);
		poperror();
	} else {
		/* failure: save the error buffer for errstr */
		e = up->syserrstr;
		up->syserrstr = up->errstr;
		up->errstr = e;
//xprint("%s:%d %s error %s\n", up->text, up->pid, sysctab[scallnr], e);
	}
	if(up->nerrlab) {
		print("bad errstack [%uld]: %d extra\n", scallnr, up->nerrlab);
		for(i = 0; i < NERR; i++)
			print("sp=%lux pc=%lux\n",
				up->errlab[i].sp, up->errlab[i].pc);
		panic("error stack");
	}

	ur->pc += 4;
	ur->npc = ur->pc + 4;
	ur->r7 = ret;

	if(up->procctl == Proc_tracesyscall){
		up->procctl = Proc_stopme;
		s = splhi();
		procctl(up);
		splx(s);
	}

	up->psstate = 0;
	up->insyscall = 0;
	if(scallnr == NOTED)
		noted(ur, *(ulong*)(sp + 1*BY2WD));

	if(scallnr != RFORK && (up->procctl || up->nnote)) {
		splhi();
		if(notify(ur))
			ret = ur->r7;
	}

	/* if we delayed scheduling, do it now */
	splhi();
	if(up->delaysched)
		sched();

//xprint("%s:%d syscall returns %x\n", up->text, up->pid, ret);
	spllo();
	kexit(ur);
	return ret;
}

/* 
 * port: Build a return frame that returns 0 from syscall().
 */
void
forkchild(Proc *p, Ureg *ur)
{
	Ureg *cur;

	/*
	 * forkret expects 8 bytes to be discarded followed
	 * by ureg.  It takes care of setting R7 to zero.
	 */
	p->sched.sp = (ulong)p->kstack+KSTACK-(2*BY2WD+sizeof(Ureg));
	p->sched.pc = (ulong)forkret - 8;  /* return to pc+8 */
	cur = (Ureg*)(p->sched.sp+2*BY2WD);
	*cur = *ur;

	/* these steps normally done at the end of syscall */
	cur->pc += 4;
	cur->npc = cur->pc + 4;
	p->psstate = 0;
	p->insyscall = 0;
}


/* port */
void
evenaddr(ulong addr)
{
	if(addr & 3) {
		postnote(up, 1, "sys: odd address", NDebug);
		error(Ebadarg);
	}
}

void
trapcritical(ulong sp, ulong pc, ulong tt)
{
	int pos;

	xprint("critical: pc %x, sp %x, tt %x\n", pc, sp, tt);
	pos = (dbgpos - 4) & (NDBG-1);
	for(; pos != dbgpos; pos = (pos-4) & (NDBG-1)) {
		xprint("%x %x %x %x\n", trapdbg[pos], trapdbg[pos+1],
			trapdbg[pos+2], trapdbg[pos+3]);
	}
	
	rdb();
}
