/*
 * Driver for the Z8530.
 * Based on older devscc.c from 9.95/port.
 */
#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"

/* #define ZDEBUG /**/

enum
{
	/* wr 0 */
	ResExtPend=	2<<3,
	ResTxPend=	5<<3,
	ResErr=		6<<3,
	ResIntr=		7<<3,

	/* wr 1 */
	ExtIntEna=	1<<0,
	TxIntEna=	1<<1,
	RxIntDis=	0<<3,
	RxIntFirstEna=	1<<3,
	RxIntAllEna=	2<<3,

	/* wr 3 */
	RxEna=		1,
	Rx5bits=	0<<6,
	Rx7bits=	1<<6,
	Rx6bits=	2<<6,
	Rx8bits=	3<<6,
	Rxbitmask=	3<<6,

	/* wr 4 */
	ParEven=	3<<0,
	ParOdd=		1<<0,
	ParOff=		0<<0,
	ParMask=	3<<0,
	SyncMode=	0<<2,
	Rx1stop=	1<<2,
	Rx1hstop=	2<<2,
	Rx2stop=	3<<2,
	X16=		1<<6,

	/* wr 5 */
	TxRTS=		1<<1,
	TxEna=		1<<3,
	TxBreak=	1<<4,
	TxDTR=		1<<7,
	Tx5bits=	0<<5,
	Tx7bits=	1<<5,
	Tx6bits=	2<<5,
	Tx8bits=	3<<5,
	Txbitmask=	3<<5,

	/* wr 9 */
	IntEna=		1<<3,
	ResetB=		1<<6,
	ResetA=		2<<6,
	HardReset=	3<<6,

	/* wr 11 */
	TRxCOutBR=	2,
	TxClockTRxC=	1<<3,
	TxClockBR=	2<<3,
	RxClockTRxC=	1<<5,
	RxClockBR=	2<<5,
	TRxCOI=		1<<2,

	/* wr 14 */
	BREna=		1,
	BRSource=	2,

	/* rr 0 */
	RxReady=	1,
	TxReady=	1<<2,
	RxDCD=		1<<3,
	RxCTS=		1<<5,
	RxBreak=	1<<7,

	/* rr 3 */
	ExtPendB=	1,	
	TxPendB=	1<<1,
	RxPendB=	1<<2,
	ExtPendA=	1<<3,	
	TxPendA=	1<<4,
	RxPendA=	1<<5,
};

extern PhysUart z8530physuart;	/* declared below */

typedef struct Ctrl {
	Lock;
	struct {
		uchar	csr;
		uchar	dummy1;
		uchar	dr;
		uchar	dummy2;
	} *ptr;				/* hardware registers */
	uchar	sticky[16];	/* sticky fields for regs */
#ifdef ZDEBUG
	uchar	last[16];
#endif
	uchar	mask;
} Ctrl;

#define NUART	4
static Uart z8530uart[NUART];
static int nuart = 0;

/*
 * helpers for accessing hardware registers.
 * Call with controller lock held.
 */
static void
csr8w(Ctrl *cp, int addr, uchar value)
{
	cp->ptr->csr = addr;
	cp->ptr->csr = cp->sticky[addr] | value;
#ifdef ZDEBUG
	cp->last[addr] = cp->sticky[addr] | value;
#endif
}

static uchar
csr8r(Ctrl *cp, int addr)
{
	cp->ptr->csr = addr;
	return cp->ptr->csr;
}

/*
 * Called for each uart phys.  Return a list of all the
 * uart devices for the phys.
 */
static Uart *
z8530pnp(void)
{
	int i;

	for(i = 0; i < nuart - 1; i++)
		z8530uart[i].next = &z8530uart[i+1];
	return &z8530uart[0];
}

/*
 *  set the baud rate by calculating and setting the baudrate
 *  generator constant.  This will work with fairly non-standard
 *  baud rates.
 */
static int
z8530baud(Uart *uart, int baud)
{
	Ctrl* ctrl;
	int brconst;

	if(baud <= 0)
		return -1;
	ctrl = uart->regs;
	brconst = (uart->freq+16*baud-1)/(2*16*baud) - 2;
	ilock(ctrl);
	csr8w(uart->regs, 12, brconst & 0xff);
	csr8w(uart->regs, 13, (brconst>>8) & 0xff);
	uart->baud = baud;
	iunlock(ctrl);
	return 0;
}

static int
z8530parity(Uart *uart, int parity)
{
	Ctrl *ctrl;
	int val;

	ctrl = uart->regs;
	switch(parity) {
	case 'e':
		val = ParEven;
		break;
	case 'o':
		val = ParOdd;
		break;
	case 'n':
		val = ParOff;
		break;
	default:
		return -1;
	}
	ilock(ctrl);
	ctrl->sticky[4] = (ctrl->sticky[4] & ~ParMask) | val;
	csr8w(ctrl, 4, 0);
	uart->parity = parity;
	iunlock(ctrl);
	return 0;
}

/*
 *  set bits/character.
 */
static int
z8530bits(Uart *uart, int n)
{
	Ctrl *ctrl;
	int rbits, tbits;

	ctrl = uart->regs;
	ilock(ctrl);
	switch(n){
	case 5:
		ctrl->mask = 0x1f;
		rbits = Rx5bits;
		tbits = Tx5bits;
		break;
	case 6:
		ctrl->mask = 0x3f;
		rbits = Rx6bits;
		tbits = Tx6bits;
		break;
	case 7:
		ctrl->mask = 0x7f;
		rbits = Rx7bits;
		tbits = Tx7bits;
		break;
	case 8:
		ctrl->mask = 0xff;
		rbits = Rx8bits;
		tbits = Tx8bits;
		break;
	default:
		return -1;
	}
	ctrl->sticky[3] = (ctrl->sticky[3]&~Rxbitmask) | rbits;
	csr8w(ctrl, 3, 0);
	ctrl->sticky[5] = (ctrl->sticky[5]&~Txbitmask) | tbits;
	csr8w(ctrl, 5, 0);
	uart->bits = n;
	iunlock(ctrl);
	return 0;
}

/*
 *  toggle DTR
 */
static void
z8530dtr(Uart *uart, int on)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	ilock(ctrl);
	if(on)
		ctrl->sticky[5] |= TxDTR;
	else
		ctrl->sticky[5] &= ~TxDTR;
	csr8w(ctrl, 5, 0);
	iunlock(ctrl);
}

/*
 *  toggle RTS
 */
void
z8530rts(Uart *uart, int on)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	ilock(ctrl);
	if(on)
		ctrl->sticky[5] |= TxRTS;
	else
		ctrl->sticky[5] &=~TxRTS;
	csr8w(ctrl, 5, 0);
	iunlock(ctrl);
}

/*
 *  send break
 */
static void
z8530break(Uart *uart, int ms)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	if(ms == 0)
		ms = 200;
	ilock(ctrl);
	ctrl->sticky[1] &=~TxIntEna;
	csr8w(ctrl, 1, 0);
	csr8w(ctrl, 5, TxBreak|TxEna);
	tsleep(&up->sleep, return0, 0, ms);
	csr8w(ctrl, 5, 0);

	/* XXX still needed?  i8250 doesnt do this */
	if(uart->oq) {
		ctrl->sticky[1] |= TxIntEna;
		csr8w(ctrl, 1, 0);
	}
	iunlock(ctrl);
}

/*
 *  set number of stop bits
 */
static int
z8530stop(Uart *uart, int stop)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	ilock(ctrl);
	ctrl->sticky[4] &=~Rx2stop;
	switch(stop) {
	case 1:
		ctrl->sticky[4] |= Rx1hstop;
		break;
	case 2:
		ctrl->sticky[4] |= Rx2stop;
		break;
	default:
		return -1;
	}
	csr8w(ctrl, 4, 0);
	uart->stop = stop;
	iunlock(ctrl);
	return 0;
}

static void
z8530kick(Uart *uart)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	ilock(ctrl);
	while(csr8r(ctrl, 0) & TxReady) {
		/* break if output queue is full */
		if(uart->op >= uart->oe && uartstageoutput(uart) == 0) {
			csr8w(ctrl, 0, ResTxPend);
			break;
		}
		ctrl->ptr->dr = *(uart->op)++;
	}
	iunlock(ctrl);
}

static void
z8530modemctl(Uart *uart, int on)
{
	/* XXX what to do? */
	USED(uart, on);
}

/*
 *  turn on a port's interrupts.  set DTR and RTS
 */
static void
z8530enable(Uart *uart, int intenb)
{
	Ctrl *ctrl;

	if(intenb) {
		ctrl = uart->regs;
		ilock(ctrl);
		ctrl->sticky[1] |= RxIntAllEna | TxIntEna ;/*XXX | ExtIntEna;*/
		csr8w(ctrl, 1, 0);
		ctrl->sticky[9] |= IntEna;
		csr8w(ctrl, 9, 0);
		iunlock(ctrl);

		/* turn on DTR and RTS */
		z8530dtr(uart, 1);
		z8530rts(uart, 1);
	}
}

static void
z8530disable(Uart *uart)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	ilock(ctrl);
	ctrl->sticky[1] &= ~(RxIntAllEna | TxIntEna | ExtIntEna);
	csr8w(ctrl, 1, 0);
	ctrl->sticky[9] &= ~IntEna;
	csr8w(ctrl, 9, 0);
	iunlock(ctrl);
}

static void
z8530fifo(Uart *uart, int)
{
	/* XXX what to do? */
	USED(uart);
}

static int
z8530getc(Uart *uart)
{
	Ctrl *ctrl;
	int ch;

	ctrl = uart->regs;
	ilock(ctrl);
	while(!(csr8r(ctrl, 0) & RxReady))
		continue;
	ch = ctrl->ptr->dr & ctrl->mask;
	iunlock(ctrl);
	if(ch == '\r') return '\n';
	return ch;
}

static void
z8530putc(Uart *uart, int c)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	ilock(ctrl);
	while(!(csr8r(ctrl, 0) & TxReady))
		continue;
	ctrl->ptr->dr = c & ctrl->mask;
	iunlock(ctrl);
}

static long
z8530status(Uart *uart, void *buf, long n, long offset)
{
	static char bittab[] = { 5, 7, 6, 8 };
	static char partab[] = { 'n', 'o', 's', 'e' };
	static char stoptab[] = { 0, 1, -1, 2 };
	Ctrl *ctrl;
	char p[128];
	char bits, par, stops;

	ctrl = uart->regs;
	bits = bittab[(ctrl->sticky[3] & Rxbitmask) >> 6];
	par = partab[(ctrl->sticky[4] & ParMask) >> 0];
	stops = stoptab[(ctrl->sticky[4] & Rx2stop) >> 2];

	snprint(p, sizeof p, "b%d c%d e%d l%d m0 p%c s%d i1\n"
		"dev(%d) type(%d) framing(%d) overruns(%d)\n",
		uart->baud,
		uart->hup_dcd, 
		uart->hup_dsr,
		bits,
		par,
		stops,
		uart->dev,
		uart->type,
		uart->ferr,
		uart->oerr 
	);
	n = readstr(offset, buf, n, p);
	free(p);

	return n;
}


/*
 * Interrupt dispatched to a particular device.
 */
static void
z8530intr0(Uart *uart, uchar x)
{
	Ctrl *ctrl;
	char ch;

	ctrl = uart->regs;
	ilock(ctrl);
	csr8w(ctrl, 0, ResIntr);
	if(x & ExtPendB)
		csr8w(ctrl, 0, ResExtPend);
	if(x & RxPendB) {
		while(csr8r(ctrl, 0) & RxReady){
			/*XXX check errors in rr1 and reset*/
			ch = ctrl->ptr->dr & ctrl->mask;
			iunlock(ctrl);
			uartrecv(uart, ch);
			ilock(ctrl);
		}
	}
	if(x & TxPendB) {
		if(uart->blocked)
			csr8w(ctrl, 0, ResTxPend);
		iunlock(ctrl);
		uartkick(uart);
	} else
		iunlock(ctrl);
}

/*
 * Interrupt for z8530 devices.
 */
int
z8530intr(void)
{
	Ctrl *ctrl1;
	uchar x;
	int i, j;

	/*
	 * Devices are paired.  The interrupt states is in RD3 of
	 * channel A (the second device).
	 */
	for(i = j = 0; i < nuart; i += 2) {
		ctrl1 = z8530uart[i+1].regs;
		ilock(ctrl1);
		x = csr8r(ctrl1, 3);
		iunlock(ctrl1);

		if(x & (ExtPendB|RxPendB|TxPendB)) {
			j++;
			z8530intr0(&z8530uart[i], x);
		}
		x = x >> 3;
		if(x & (ExtPendB|RxPendB|TxPendB)) {
			j++;
			z8530intr0(&z8530uart[i+1], x);
		}
	}
	return j;
}


/*
 * Internal helper to initialize our device.
 *  default is 9600 baud, 1 stop bit, 8 bit chars, no interrupts,
 *  transmit and receive enabled, interrupts disabled.
 */
static void
z8530setup0(Uart *uart, int brsource)
{
	Ctrl *ctrl;

	ctrl = uart->regs;
	memset(ctrl->sticky, 0, sizeof(ctrl->sticky));

	/*
	 *  turn on baud rate generator and set rate to 9600 baud.
	 *  use 1 stop bit.
	 * XXX this seems overly complicated.  We should
	 * be using more of our helper functions here.
	 */
	ctrl->sticky[14] = brsource ? BRSource : 0;
	csr8w(ctrl, 14, 0);
	z8530baud(uart, 9600);
	ctrl->sticky[4] = Rx1stop | X16;
	csr8w(ctrl, 4, 0);
	ctrl->sticky[11] = TxClockBR | RxClockBR | TRxCOutBR /*| TRxCOI*/;
	csr8w(ctrl, 11, 0);
	ctrl->sticky[14] = BREna;
	if(brsource)
		ctrl->sticky[14] |= BRSource;
	csr8w(ctrl, 14, 0);

	/*
	 *  enable I/O, 8 bits/character
	 */
	ctrl->sticky[3] = RxEna | Rx8bits;
	csr8w(ctrl, 3, 0);
	ctrl->sticky[5] = TxEna | Tx8bits;
	csr8w(ctrl, 5, 0);
	ctrl->mask = 0xff;
}

/*
 * Add an 8530 controller.
 */
static void
z8530setup(void *base, ulong freq, int brsource, char *name0, char *name1)
{
	Ctrl *ctrl;
	Uart *uart;

	if(nuart + 2 > NUART)
		panic("too many z8530s");

	ctrl = xalloc(sizeof *ctrl);
	ctrl->ptr = base;
	base = ctrl->ptr + 1;
	uart = &z8530uart[nuart++];
	uart->regs = ctrl;
	uart->name = name0;
	uart->freq = freq;
	uart->phys = &z8530physuart;
	z8530setup0(uart, brsource);

	ctrl = xalloc(sizeof *ctrl);
	ctrl->ptr = base;
	uart = &z8530uart[nuart++];
	uart->regs = ctrl;
	uart->name = name1;
	uart->freq = freq;
	uart->phys = &z8530physuart;
	z8530setup0(uart, brsource);
}


PhysUart z8530physuart = {
	.name	= "z8530",
	.pnp		= z8530pnp,
	.enable	= z8530enable,
	.disable	= z8530disable,
	.kick		= z8530kick,
	.dobreak	= z8530break,
	.baud	= z8530baud,
	.bits		= z8530bits,
	.stop		= z8530stop,
	.parity	= z8530parity,
	.modemctl = z8530modemctl,
	.rts		= z8530rts,
	.dtr		= z8530dtr,
	.status	= z8530status,
	.fifo		= z8530fifo,
	.getc		= z8530getc,
	.putc	= z8530putc,
};


/* initialize z8530 drivers and serial console for sbus based sparc */
void
z8530init(void)
{
	KMap *k;
	Uart *uart;
	char *cmd, *p;
	ulong n;

	/* setup our four serial devices */
	k = kmapphys(KMDUART, 1, PTEIO|PTEUNCACHED|PTELOCK);
	z8530setup((void*)VA(k), KMFREQ, 1, "mouse", "keyboard");
	k = kmapphys(EIADUART, 1, PTEIO|PTEUNCACHED|PTELOCK);
	z8530setup((void*)VA(k), EIAFREQ, 1, "ttyb", "ttya");
	sbusenableintr(SbusIntrMapSerial);

	/* setup one of our devices as the console */
	if((p = getconf("console")) == nil)
		p = "3 b9600 l8 pn s1";
	n = strtoul(p, &cmd, 0);
	if(p != cmd && n < nuart) {
		uart = z8530uart + n;
		(*uart->phys->enable)(uart, 0);
		if(*cmd)
			uartctl(uart, cmd);
		uart->console = 1;
		consuart = uart;
	}
	xputcfunc = uartputc;
}

/* DEBUG */
void
dumpuart(Uart *u)
{
	unsigned char buf[16];
	Ctrl *ctrl;
	int i;

	ctrl = u->regs;
	for(i = 0; i < 16; i++)
		buf[i] = csr8r(ctrl, i);
	xprint("dump uart %s: ", u->name);
	for(i = 0; i < 16; i++)
		xprint("%x ", buf[i]);
	xprint("\n");
#ifdef ZDEBUG
	for(i = 0; i < 16; i++)
		xprint("%x ", ctrl->last[i]);
	xprint("\n");
#endif
}

